"""
PyRasgo Custom Errors
"""
from typing import List


### BASELINE RASGO EXCEPTION
class PyRasgoException(Exception):
    """
    Base error attributable to PyRasgo code
    """

    def __init__(self, message: any):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


### API & PACKAGE EXCEPTIONS
class APIError(PyRasgoException):
    """
    Error message returned from Rasgo API
    """


class APITimeoutWarning(APIError):
    """
    API Request Timed Out :(
    """


class PackageDependencyWarning(PyRasgoException):
    """
    Problem importing a required package dependency
    """


class ParameterValueError(PyRasgoException):
    """
    Incorrect parameter passed into a pyrasgo function
    """

    def __init__(self, param: str = None, values: List[str] = None, message: str = None):
        if not message:
            message = f"Please pass a valid value for {param}. Values are {values}"
        super().__init__(message=message)
        self.message = message


### RASGO LOGICAL EXCEPTIONS
class RasgoResourceException(PyRasgoException):
    """
    Cannot access this resource
    """


class RasgoRuleViolation(PyRasgoException):
    """
    Business rule violated
    """


class RasgoSchemaException(PyRasgoException):
    """
    Schema expectation violated
    """


class TransformRenderingException(PyRasgoException):
    """
    Error from Transform rendering activity
    """


### DATAWAREHOUSE EXCEPTIONS
class DWCredentialsWarning(PyRasgoException):
    """
    Missing DataWarehouse credentials
    """


class DWQueryException(PyRasgoException):
    """
    Error from DataWarehouse query activity
    """


class DWResourceException(PyRasgoException):
    """
    Error accessing Table
    """



import logging
import webbrowser

from pyrasgo import errors
from pyrasgo.config import set_session_api_key
from pyrasgo.version import __version__

__all__ = [
    '__version__',
    'connect',
    'open_docs',
    'pronounce_rasgo',
    'login',
]


def connect(api_key):
    from pyrasgo.rasgo import Rasgo

    set_session_api_key(api_key)
    return Rasgo()


def open_docs():
    webbrowser.open("https://docs.rasgoml.com/rasgo-docs/reference/pyrasgo")


def pronounce_rasgo():
    webbrowser.open("https://www.spanishdict.com/pronunciation/rasgo?langFrom=es")


def login(email: str, password: str):
    from pyrasgo.api.login import Login
    from pyrasgo.schemas.user import UserLogin

    payload = UserLogin(
        email=email,
        password=password,
    )
    try:
        response = Login().login(payload=payload)
        return connect(api_key=response)
    except Exception as err:
        raise errors.APIError("Unable to log in with credentials provided") from err



"""
API Update Functions
"""
import time
from typing import Dict, List, Optional, Union

try:
    from typing import Literal  # Only available on 3.8+ directly
except ImportError:
    from typing_extensions import Literal

from pyrasgo import errors, primitives, schemas
from pyrasgo.config import MAX_POLL_ATTEMPTS, POLL_RETRY_RATE


class Update:
    """
    API Update Class
    """

    def __init__(self):
        from pyrasgo.api import Get
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()
        self._dw = None

    def dataset(
        self,
        dataset: primitives.Dataset,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        attributes: Optional[dict] = None,
    ) -> primitives.Dataset:
        """
        Update a dataset name, description, and/or attributes in Rasgo
        """
        # Raise error if trying to update a dataset in offline mode
        if not dataset._api_dataset:
            raise errors.RasgoRuleViolation(
                "Can not update dataset. Needs to be saved first with `rasgo.save.dataset()`"
            )

        dataset_update = schemas.DatasetUpdate(
            # Possible Changed Fields
            name=name,
            description=description,
            # Persist other fields in update contract so no fields set to None in update
            status=dataset._api_dataset.status,
            owner_id=dataset._api_dataset.owner_id,
            dw_table_id=dataset._api_dataset.dw_table_id,
            attributes=attributes,
        )
        response = self.api._put(
            f"/datasets/{dataset._api_dataset.id}",
            dataset_update.dict(exclude_unset=True, exclude_none=True),
            api_version=2,
        ).json()
        dataset_schema = schemas.Dataset(**response)
        return primitives.Dataset(api_dataset=dataset_schema)

    def dataset_column(
        self,
        dataset_column_id: int,
        display_name: Optional[str] = None,
        description: Optional[str] = None,
        attributes: Optional[dict] = None,
        tags: Optional[List[str]] = None,
    ) -> schemas.DatasetColumn:
        """
        Update metadata about a dataset column

        Args:
            dataset_column_id: Dataset column id to updated. Use `dataset.columns[x].id` to retrieve
            display_name: Display name to update for this dataset column if set
            description: Description to update for this dataset column if set
            attributes: Attributes to add or update for this dataset column. Set as Key Value pairs dict
            tags: Tags to add to this dataset column

        Returns:
            Updated Dataset Column Obj
        """
        ds_col_update = schemas.DatasetColumnUpdate(
            display_name=display_name,
            description=description,
            attributes=attributes,
            tags=tags,
        )
        resp = self.api._put(
            f"/dataset-columns/{dataset_column_id}",
            ds_col_update.dict(exclude_unset=True, exclude_none=True),
            api_version=2,
        ).json()
        return schemas.DatasetColumn(**resp)

    def dataset_table(self, dataset: primitives.Dataset, verbose: bool = False, timeout: Optional[int] = None) -> None:
        """
        Kicks off a query for re-materializing a dataset's set Dw Table.

        The dataset needs to be published first in order call this function

        Args:
            dataset: Published dataset to refresh table for
            verbose: If True will print information related to refreshing table
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
        """
        # We need to ensure that the API dataset is exists and is published
        if not dataset._api_dataset or dataset.status.lower() != "published":
            raise errors.RasgoRuleViolation('Can not refresh table. Dataset must first be published')

        if verbose:
            print(f"Refreshing table for dataset with id '{dataset.id}' at fqtn: '{dataset.fqtn}'")

        # Call endpoint to kick off Query to re-create table in worker
        response = self.api._put(f"/datasets/{dataset._api_dataset.id}/table-refresh", api_version=2).json()
        status_tracking = schemas.StatusTracking(**response)

        # Poll the async job until completion, failure, or timeout
        for i in range(1, MAX_POLL_ATTEMPTS):
            status_tracking = schemas.StatusTracking(
                **self.api._get(f"/status-tracking/{status_tracking.tracking_uuid}", api_version=2).json()
            )
            if status_tracking.status == "completed":
                if verbose:
                    print(f"Done Refreshing table for dataset with id '{dataset.id}' " f"at fqtn: '{dataset.fqtn}'")
                break
            if status_tracking.status == "failed":
                raise errors.APIError(f"Could not refresh dataset table: {status_tracking.message}")
            if timeout and (POLL_RETRY_RATE * i) > timeout:
                raise errors.APITimeoutWarning("Timeout reached waiting for dataset table refreshing")
            time.sleep(POLL_RETRY_RATE)
            if verbose:
                print("Query still executing...")

    def dataset_schema(self, dataset_id: int) -> None:
        """
        Queries the information_schema to get column metadata from Snowflake
        Then updates this dataset's column metadata in Rasgo to match
        """
        self.api._put(
            f"/datasets/{dataset_id}/sync-schema",
            api_version=2,
        ).json()

    def metric(
        self,
        metric_id: int,
        dataset_id: Optional[int] = None,
        name: Optional[str] = None,
        aggregation_type: Optional[str] = None,
        target_expression: Optional[str] = None,
        time_grains: Optional[List[Literal["HOUR", "DAY", "WEEK", "MONTH", "QUARTER", "YEAR"]]] = None,
        time_dimension: Optional[str] = None,
        dimensions: Optional[List[str]] = None,
        filters: Optional[List[schemas.Filter]] = None,
        meta: Optional[Dict[str, str]] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> schemas.Metric:
        """
        Updates a metric on a dataset
        """
        time_grain_objs = None
        if time_grains:
            time_grain_objs = [schemas.metric.TimeGrain[grain] for grain in time_grains]
        metric = schemas.MetricUpdate(
            id=metric_id,
            ds_dataset_id=dataset_id,
            name=name,
            aggregation_type=aggregation_type,
            target_expression=target_expression,
            time_grains=time_grain_objs,
            time_dimension=time_dimension,
            dimensions=dimensions,
            filters=filters,
            meta=meta,
            label=label,
            description=description,
        )
        try:
            response = self.api._put(
                f"/metric", metric.dict(exclude_unset=True, exclude_none=True), api_version=2
            ).json()
            return schemas.Metric(**response)
        except Exception as err:
            raise errors.RasgoResourceException(f"Could not update metric with ID '{metric_id}'.") from err

    def transform(
        self,
        transform_id: int,
        name: Optional[str] = None,
        source_code: Optional[str] = None,
        type: Optional[str] = None,
        arguments: Optional[List[dict]] = None,
        description: Optional[str] = None,
        tags: Optional[Union[List[str], str]] = None,
        dw_type: Optional[Literal["SNOWFLAKE", "BIGQUERY", "UNSET"]] = None,
    ) -> schemas.Transform:
        """
        Updates a transform in Rasgo

        Args:
            transform_id: Id of transform to update
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform
            tags: List of tags, or a tag (string), to set on this dataset
            dw_type: DataWarehouse provider: SNOWFLAKE, BIGQUERY or UNSET
        """
        # Init tag array to be list of strings
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        # Make request to update transform and return
        transform_update = schemas.TransformUpdate(
            name=name,
            type=type,
            description=description,
            sourceCode=source_code,
            arguments=arguments,
            tags=tags,
            dw_type=dw_type.upper() if dw_type else None,
        )
        response = self.api._put(
            f"/transform/{transform_id}",
            transform_update.dict(exclude_unset=True, exclude_none=True),
            api_version=1,
        ).json()
        return schemas.Transform(**response)



"""
API Get Functions
"""
from typing import List, Optional

from pyrasgo import errors, primitives, schemas


class Get:
    """
    API Get Class
    """

    def __init__(self):
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self._dw = None

    def accelerator(self, accelerator_id: int) -> schemas.Accelerator:
        """
        Return an Accelerator with specified id
        """
        response = self.api._get(f"/accelerators/{accelerator_id}", api_version=2).json()
        return schemas.Accelerator(**response)

    def accelerators(self) -> List[schemas.Accelerator]:
        """
        Return a list of all Accelerator's available to user
        """
        # Make API call get get bulk Accelerator Json objs
        response = self.api._get("/accelerators", api_version=2).json()

        # Convert response to Accelerator Primitives
        accelerators = []
        for accelerator_json in response:
            api_accelerator = schemas.Accelerator(**accelerator_json)
            accelerators.append(api_accelerator)
        return accelerators

    def dataset(
        self,
        dataset_id: Optional[int] = None,
        fqtn: Optional[str] = None,
        resource_key: Optional[str] = None,
        version: Optional[int] = None,
    ) -> primitives.Dataset:
        """
        Return a Rasgo dataset primitive by Id or FQTN
        """
        no_input_supplied = dataset_id is None and not fqtn and not resource_key
        too_many_inputs_supplied = (bool(dataset_id) + bool(fqtn) + bool(resource_key or version)) > 1
        if no_input_supplied or too_many_inputs_supplied:
            raise errors.ParameterValueError(
                message="Valid Dataset retrieval options are:\n1. id\n2. fqtn\n3. resource_key\n4. resource_key and version"
            )

        try:
            if dataset_id:
                response = self.api._get(f"/datasets/{dataset_id}", api_version=2).json()
            elif fqtn:
                response = self.api._get(f"/datasets/match/{fqtn}", api_version=2).json()
            elif resource_key:
                response = self.api._get(
                    f"/datasets/rk/{resource_key}{f'/versions/{version}' if version else ''}",
                    api_version=2,
                ).json()
            dataset_schema = schemas.Dataset(**response)

            operation_set_schema = None
            if dataset_schema.dw_operation_set_id:
                response = self.api._get(
                    f"/operation-sets/{dataset_schema.dw_operation_set_id}",
                    api_version=2,
                ).json()
                operation_set_schema = schemas.OperationSet(**response)
            return primitives.Dataset(
                api_dataset=dataset_schema,
                api_operation_set=operation_set_schema,
            )
        except Exception:
            if dataset_id:
                raise errors.RasgoResourceException(
                    f"Dataset with id '{dataset_id!r}' does not exist or this API key does not have access."
                )
            elif fqtn:
                raise errors.RasgoResourceException(
                    f"Dataset with fqtn '{fqtn}' does not exist or this API key does not have access."
                )

    def datasets(
        self,
        published_only: bool = False,
        include_community: bool = False,
    ) -> List[primitives.Dataset]:
        """
        Return all datasets in Rasgo attached to your organization

        Params
        `only_published` boolean:
            Instructs whether to return all datasets (including drafts) or
            only published datasets
        """
        # Get transforms so we can cache them for use in transforming datasets
        transforms = self.transforms()
        PAGE_SIZE = 100
        offset = 0

        datasets = []
        while True:
            response = self.api._get(
                f"/datasets?page_size={PAGE_SIZE}&page_start={offset}"
                f"&include_community={include_community}"
                f"&published_only={published_only}",
                api_version=2,
            ).json()
            for r in response:
                dataset_schema = schemas.Dataset(**r)
                dataset = primitives.Dataset(
                    api_dataset=dataset_schema,
                    transforms=transforms,
                )
                datasets.append(dataset)
            if len(response) < PAGE_SIZE:
                break
            offset += PAGE_SIZE
        return datasets

    def dataset_metrics(self, dataset_id: int) -> List[schemas.Metric]:
        """
        Return a list of metrics belonging to a dataset
        """
        response = self.api._get(f"/datasets/{dataset_id}/metrics", api_version=2).json()
        return [schemas.Metric(**metric) for metric in response]

    def dataset_offline_version(
        self,
        resource_key: Optional[str],
        version: Optional[int] = None,
    ) -> schemas.OfflineDataset:
        try:
            target = f"/datasets/{resource_key}/offline-version"
            if version:
                target = f"{target}?version={version}"
            response = self.api._get(resource=target, api_version=2).json()
            return schemas.OfflineDataset(**response)
        except Exception as err:
            raise errors.RasgoResourceException(
                f"Dataset with key '{resource_key!r}' does not exist or this API key does not have access."
            ) from err

    def dataset_py(self, dataset_id: int) -> str:
        """
        Return the pyrasgo code which will create an offline copy
        of a dataset (by ds id)  whether DS is in draft or
        unpublished status
        """
        resp = self.api._get(
            f"/datasets/{dataset_id}/export/python",
            api_version=2,
        ).json()
        return resp

    def metrics(self, dataset_id: int) -> List[schemas.Metric]:
        """
        Return a list of metrics belonging to a dataset
        """
        return self.dataset_metrics(dataset_id)

    def transform(self, transform_id: int) -> schemas.Transform:
        """Returns an individual transform"""
        try:
            response = self.api._get(f"/transform/{transform_id}", api_version=1).json()
            return schemas.Transform(**response)
        except Exception as err:
            raise errors.RasgoResourceException(
                f"Transform with id '{transform_id}' does not exist or this API key does not have access."
            ) from err

    def transforms(self) -> List[schemas.Transform]:
        """Returns a list of available transforms"""
        response = self.api._get("/transform", api_version=1).json()
        return [schemas.Transform(**r) for r in response]

    def community_transforms(self) -> List[schemas.Transform]:
        response = self.api._get("/transform/community", api_version=1).json()
        return [schemas.Transform(**r) for r in response]

    def user(self):
        response = self.api._get("/users/me", api_version=1).json()
        return schemas.User(**response)

    # ----------------------------------
    #  Internal/Private Get Calls
    # ----------------------------------

    def _dataset_columns(self, dataset_id: int) -> List[schemas.DatasetColumn]:
        """
        Return the dataset columns for a specific dataset

        Args:
            dataset_id: Id of dataset to retrieve columns for
        """
        response = self.api._get(
            f"/dataset-columns/ds/{dataset_id}",
            api_version=2,
        ).json()
        return [schemas.DatasetColumn(**x) for x in response]

    def _operation_set(self, operation_set_id: int) -> schemas.OperationSet:
        """
        Return a Rasgo operation set by id
        """
        response = self.api._get(
            f"/operation-sets/{operation_set_id}",
            api_version=2,
        ).json()
        return schemas.OperationSet(**response)

    def _operation_set_async_status(self, task_id: int) -> schemas.OperationSetAsyncTask:
        """
        Returns the status of an Operation Set creation task by id
        """
        response = self.api._get(
            f"/operation-sets/async/{task_id}",
            api_version=2,
        ).json()
        return schemas.OperationSetAsyncTask(**response)



from enum import Enum


class DataType(Enum):
    SMALLINT = 'smallint'
    BIGINT = 'bigint'
    INT = 'int'
    INTEGER = 'integer'
    TINYINT = 'tinyint'
    BYTEINT = 'byteint'
    FLOAT = 'float'
    FLOAT4 = 'float4'
    FLOAT8 = 'float8'
    FLOAT64 = 'float64'
    DECIMAL = 'decimal'
    NUMERIC = 'numeric'
    NUMBER = 'number'
    REAL = 'real'
    DOUBLE = 'double'
    STRING = 'string'
    TEXT = 'text'
    VARCHAR = 'varchar'
    CHAR = 'char'
    CHARACTER = 'character'
    DATE = 'date'
    DATETIME = 'datetime'
    TIME = 'time'
    TIMESTAMP = 'timestamp'
    TIMESTAMP_LTZ = 'timestamp_ltz'
    TIMESTAMP_NTZ = 'timestamp_ntz'
    TIMESTAMP_TZ = 'timestamp_tz'
    BINARY = 'binary'
    VARBINARY = 'varbinary'
    BOOLEAN = 'boolean'
    BOOL = 'bool'
    VARIANT = 'variant'
    OBJECT = 'object'
    ARRAY = 'array'



from pyrasgo.api.create import Create
from pyrasgo.api.delete import Delete
from pyrasgo.api.get import Get
from pyrasgo.api.login import Login
from pyrasgo.api.publish import Publish
from pyrasgo.api.read import Read
from pyrasgo.api.update import Update



"""
BigQuery DataWarehouse Class
"""
from typing import Dict, List, Optional, Tuple
from typing_extensions import Literal

import pandas as pd

from pyrasgo.api.session import Session
from pyrasgo import errors
from pyrasgo.imports import bq, gcp_exc, gcp_flow, gcp_svc
from pyrasgo.storage.dataframe import utils as dfutils
from pyrasgo.storage.datawarehouse import utils as dwutils


class BigQueryDataWarehouse(Session):
    """
    BigQuery DataWarehouse Class
    """

    def __init__(self):
        if bq is None:
            raise errors.PackageDependencyWarning(
                "Missing a required python package to run BigQuery. "
                "Please download the BigQuery package by running: "
                "pip install pyrasgo[bigquery]"
            )
        _creds = self.profile.get("dw_creds")
        if not _creds:
            raise errors.DWCredentialsWarning("Your user is missing credentials, please contact Rasgo.")
        self._connection: bq.Client = None
        self._credentials: bq.Credentials = None
        self._key = _creds.get("json_key")
        self.default_database = _creds.get("project")
        self.default_schema = _creds.get("dataset")

    @property
    def connection(self) -> 'bq.Client':
        """
        Returns a connection to the BQ Client
        """
        if self._connection:
            return self._connection
        if not self._credentials:
            self.set_credentials()
        return bq.Client(credentials=self._credentials, project=self.default_database)

    def set_credentials(self):
        """
        Sets the credentials for this BQ Class instance
        """
        # Try to auth as a user first, then try as a service account if that fails
        try:
            self._credentials = self._get_appflow_credentials()
        # If this is a service account, BQ will raise:
        # ValueError: Client secrets must be for a web or installed app.
        except ValueError:
            self._credentials = self._get_service_account_credentials()

    def _get_service_account_credentials(self):
        """
        Google stuff
        """
        return gcp_svc.Credentials.from_service_account_info(
            self._key, scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )

    def _get_appflow_credentials(self):
        """
        Google stuff
        """
        appflow = gcp_flow.InstalledAppFlow.from_client_config(
            self._key, scopes=["https://www.googleapis.com/auth/bigquery"]
        )
        appflow.run_local_server()
        # appflow.run_console()
        return appflow.credentials

    def _job_config(self):
        """
        Returns a basic job config to run queries in default project & dataset
        """
        return bq.QueryJobConfig(default_dataset=f"{self.default_database}.{self.default_schema}")

    def execute_query(
        self,
        query: str,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> List[Tuple]:
        """
        Execute a query against this DataWarehouse and return results as a list.

        Args:
            query: SQL string to be executed

        Returns:
            List of Tuples
        """
        return list(self.connection.query(query, job_config=self._job_config()).result())

    def get_schema(
        self,
        fqtn_or_sql: str,
    ) -> Dict[str, str]:
        """
        Return the schema of a table, view, or select statement

        Params:
        `fqtn_or_sql`: str:
            Either a Fully-qualified table name (database.schema.table)
            or a SQL select statement that will create a view.
        """
        # Check for SQL
        if dwutils.is_select_statement(fqtn_or_sql):
            query_job = self.connection.query(
                query=fqtn_or_sql,
                job_config=bq.QueryJobConfig(dry_run=True),
            )
            schema = query_job._properties['statistics']['query']['schema']['fields']
            return {schema_field['name']: schema_field['type'] for schema_field in schema}
        # Otherwise assume fqtn:
        table = self.connection.get_table(fqtn_or_sql)
        schema = table.schema
        return {schema_field.name: schema_field.field_type for schema_field in schema}

    def query_into_dict(
        self,
        query: str,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> List[Dict]:
        """
        Execute a query against this DataWarehouse and return results as a dictionary.

        Args:
            query: SQL string to be executed

        Returns:
            List of Dicts
        """
        return list(self.connection.query(query, job_config=self._job_config()).result())

    def query_into_dataframe(
        self,
        query: str,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> pd.DataFrame:
        """
        Execute a query against this DataWarehouse and return results as a pandas DataFrame.

        Args:
            query: SQL select statement to be executed

        Returns:
            pandas DataFrame
        """
        return self.connection.query(query, job_config=self._job_config()).result().to_dataframe()

    def list_table_columns(
        self,
        fqtn: str,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> List[Tuple[str, str]]:
        """
        Return a list of columns in a table or view

        Args:
            fqtn: Fully-qualified table name (database.schema.table)

        Returns:
            List of Tuples
        """
        table = self.connection.get_table(fqtn)
        return [(x.name, x.field_type) for x in table.schema]

    def make_select_statement(
        self,
        table_metadata: Dict[str, str],
        filters: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
        limit: Optional[int] = None,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> str:
        """
        Constructs & returns a SQL select statement

        Args:
            table_metadata: Dict must include fqtn, or database + schema + table
            filters: List of SQL WHERE filters strings to filter on in query
            order_by: List of columns to order by in query
            columns: List of columns to return in the query
            limit: Limit the number of rows returned in the query

        Returns:
            string
        """
        # Get table fqtn
        fqtn = table_metadata.get('fqtn')
        if not fqtn:
            fqtn = "{database}.{schema}.{table}".format(**table_metadata)

        # Create Initial select statement
        columns = ', '.join(columns) if columns else '*'
        query = f"SELECT {columns} FROM {fqtn}"

        # Add filters, orders, and limit if supplied
        if filters:
            query = f"{query} WHERE {' AND '.join(filters)}"
        if order_by:
            query = f"{query} ORDER BY {', '.join(order_by)}"
        if isinstance(limit, int):
            query = f"{query} LIMIT {limit}"
        return query.strip()

    def write_dataframe_to_table(
        self,
        df: pd.DataFrame,
        table_name: str,
        method: Literal["append", "overwrite", "fail"] = "fail",
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> None:
        """
        Uploads a pandas DataFrame into this DataWarehouse as a table

        Args:
            df: pandas DataFrame to upload
            table_name: name for the target table in the DataWarehouse
            method: Literal:
                append = if table_name already exists, append to it
                overwrite = if table_name already exists, overwrite it
                fail = if table_name already exists, fail

        Returns:
            None
        """
        # Upload the df into it
        dfutils.cleanse_sql_dataframe(df)
        write_disposition = "WRITE_EMPTY"
        if method == "append":
            write_disposition = "WRITE_APPEND"
        if method == "overwrite":
            write_disposition = "WRITE_TRUNCATE"
        job_config = bq.LoadJobConfig(write_disposition=write_disposition)
        self.connection.load_table_from_dataframe(df, table_name, job_config=job_config).result()



"""
API Contracts for Data Warehouse Operations
"""
from typing import Dict, List, Optional, Any

from pydantic import BaseModel, Field

from pyrasgo.schemas import dw_table as dw_table_schemas


class OperationCreate(BaseModel):
    """
    Contract to create an Operation on set
    """

    operation_name: str = Field(alias='operationName')
    operation_args: Dict = Field(alias='operationArgs')
    transform_id: Optional[int] = Field(alias='transformId')
    transform_name: Optional[str] = Field(alias='transformName')
    table_dependency_names: Optional[List[str]] = Field(alias="tableDependencyNames")  # DW Table names
    table_dependency_ids: Optional[List[int]] = Field(alias="tableDependencyIds")  # DW Table ids
    table_name: Optional[str] = Field(alias="tableName")

    def __eq__(self: 'OperationCreate', other: Any) -> bool:
        """
        Comparison to determine if an OperationCreate is same as another

        Returns True if Equals; False Otherwise
        """
        # If the other class isn't a OperationCreate return comparison error
        if not isinstance(other, OperationCreate):
            return False
        return (
            self.operation_name == other.operation_name
            and self.operation_args == self.operation_args
            and self.transform_id == other.transform_id
            and self.table_dependency_names == other.table_dependency_names
            and self.table_dependency_ids == other.table_dependency_ids
            and self.table_name == other.table_name
        )

    class Config:
        allow_population_by_field_name = True


class OperationUpdate(BaseModel):
    """
    Contract to update an Operation on set
    """

    operation_name: Optional[str] = Field(alias='operationName')
    operation_args: Optional[Dict] = Field(alias='operationArgs')
    transform_id: Optional[int] = Field(alias='transformId')
    table_dependency_names: Optional[List[str]] = Field(alias="tableDependencyNames")  # DW Table names
    table_dependency_ids: Optional[List[int]] = Field(alias="tableDependencyIds")  # DW Table ids
    dw_operation_set_id: Optional[int] = Field(alias='dwOperationSetId')

    class Config:
        allow_population_by_field_name = True


class Operation(BaseModel):
    """
    Contract representing a Operation
    """

    id: int
    operation_name: str = Field(alias='operationName')
    operation_args: Any = Field(alias='operationArgs')
    operation_sql: str = Field(alias='operationSQL')
    transform_id: int = Field(alias='transformId')
    dw_table_id: int = Field(alias='dwTableId')
    dw_operation_set_id: int = Field(alias='dwOperationSetId')
    dw_table: Optional[dw_table_schemas.DataTableWithColumns] = Field(alias='dataTable')
    table_dependencies: Optional[List[int]] = Field(alias="tableDependencies")  # DW Table ids

    class Config:
        allow_population_by_field_name = True


class OperationAdjacency(BaseModel):
    """
    Contract representing a Operation Adjacency
    """

    direction: str
    dw_operation_id: int = Field(alias='operationId')
    dw_table_id: int = Field(alias='dwTableId')



from __future__ import annotations
from typing import Dict, List, Tuple, TYPE_CHECKING
import re

import yaml

from pyrasgo import errors
from pyrasgo.schemas.transform import Transform

if TYPE_CHECKING:
    from pyrasgo.primitives.dataset import Dataset


def replace_source_variable_args(
    operation_args: Dict[str, any],
    used_transform: Transform,
    top_source_table_arguments: List[str],
    fqtn_to_arg_names: Dict[str, any],
) -> Tuple[any, List[any], Dict[str, str]]:
    """
    Replace source table and operation FQTNs with variable names
    """

    def replace_table_arg(
        value: str, fqtn_to_arg_names: Dict[str, str], top_source_table_arguments: List[any]
    ) -> Tuple[Dict[str, str], List[any], str]:
        """
        If this argument has a mapping to a previous operation or table we already know we need, replace the value
        with that argument name. Otherwise add a new source_table argument for it.

        returns:
            - updated fqtn mapping
            - updated top level source args
            - the replaced name of the source arg
        """
        if str(value) in fqtn_to_arg_names:
            return fqtn_to_arg_names, top_source_table_arguments, f"{{{{{fqtn_to_arg_names[value]}}}}}"

        table_arg_name = f'source_table{"_"+str(len(top_source_table_arguments)) if top_source_table_arguments else ""}'
        argument_create = {
            "name": table_arg_name,
            "description": 'A source table argument that must be passed to this Accelerator',
            "argument_type": 'dataset',
        }

        fqtn_to_arg_names[value] = table_arg_name
        top_source_table_arguments.append(argument_create)
        return fqtn_to_arg_names, top_source_table_arguments, f"{{{{{table_arg_name}}}}}"

    transform_args = {}
    for arg, value in operation_args.items():
        arg_type = next(iter([x.type for x in used_transform.arguments if x.name == arg]), None)

        if arg == 'source_table' or arg_type == 'table':
            fqtn_to_arg_names, top_source_table_arguments, replaced_value = replace_table_arg(
                value, fqtn_to_arg_names, top_source_table_arguments
            )
            transform_args[arg] = replaced_value
        elif arg_type == 'table_list' and isinstance(value, List):
            val_list = []
            for table_arg in value:
                fqtn_to_arg_names, top_source_table_arguments, replaced_value = replace_table_arg(
                    table_arg, fqtn_to_arg_names, top_source_table_arguments
                )
                val_list.append(replaced_value)
            transform_args[arg] = val_list
        else:
            transform_args[arg] = value

    return (
        {
            'transform_name': used_transform.name,
            'transform_arguments': transform_args,
        },
        top_source_table_arguments,
        fqtn_to_arg_names,
    )


def escape_name(name: str) -> str:
    """
    Remove non-alphanumerics, replace spaces with '_' and lowercase everything to get valid python var name
    """
    escaped = re.sub(r'[^a-zA-Z0-9_]', '', name).replace(' ', '_').lower()
    return f'_{escaped}' if escaped and escaped[0].isnumeric() else escaped  # leading digits bad


def dataset_to_accelerator_yaml(dataset: Dataset) -> str:

    from pyrasgo.api.get import Get

    get = Get()

    top_source_table_arguments = []
    accelerator_operations = {}
    used_transforms = {x.id: x for x in dataset._available_transforms}
    fqtn_to_arg_names = {}

    # Operations
    for operation in dataset._api_operation_set.operations:

        # get the transform def, fetch it if we haven't already got it from the list we inherited from the dataset.
        # It could be deleted and that's okay, we really just need the name.
        used_transforms[operation.transform_id] = used_transforms.get(
            operation.transform_id, get.transform(operation.transform_id)
        )
        used_transform = used_transforms.get(operation.transform_id)
        if not used_transform:
            raise errors.RasgoResourceException(
                f"Transform {operation.transform_id} referenced in {operation.operation_name} is not available"
            )

        (
            accelerator_operations[escape_name(operation.operation_name)],
            top_source_table_arguments,
            fqtn_to_arg_names,
        ) = replace_source_variable_args(
            operation.operation_args, used_transform, top_source_table_arguments, fqtn_to_arg_names
        )
        fqtn_to_arg_names[operation.dw_table.fqtn] = escape_name(operation.operation_name)

    # Insights
    for insight in dataset._api_dataset.insights:
        used_transforms[insight.transform_id] = used_transforms.get(
            insight.transform_id, get.transform(insight.transform_id)
        )
        used_transform = used_transforms.get(insight.transform_id)
        if not used_transform:
            raise errors.RasgoResourceException(
                f"Transform {insight.transform_id} referenced in {insight.name} is not available"
            )

        parsed, top_source_table_arguments, fqtn_to_arg_names = replace_source_variable_args(
            insight.transform_arguments, used_transform, top_source_table_arguments, fqtn_to_arg_names
        )
        accelerator_operations[escape_name(insight.name)] = {'operation_type': 'INSIGHT', **parsed}

    final_args = {x.pop('name'): x for x in top_source_table_arguments}
    accelerator_dict = {
        'name': f"{dataset.name} Accelerator",
        'description': f"Accelerator template created from Dataset {dataset.id}",
        'arguments': final_args,
        'operations': accelerator_operations,
    }

    # Doc section
    # Uses the output table of the dataset to build a default doc section
    output_operation = next(
        iter([x for x in dataset._api_operation_set.operations if x.dw_table_id == dataset._api_dataset.dw_table_id]),
        None,
    )
    if output_operation:
        accelerator_dict['doc'] = {
            'output_tables': {
                escape_name(output_operation.operation_name): {
                    x.column_name: {'description': f'{x.column_name}:{x.data_type}'}
                    for x in output_operation.dw_table.columns
                }
            }
        }

    return yaml.safe_dump(accelerator_dict, sort_keys=False)



"""
API Login Functions
"""
import requests

from pyrasgo import errors
from pyrasgo.api.session import Environment
from pyrasgo.schemas.user import UserLogin


class Login:
    """
    API Login Class
    """

    def login(self, payload: UserLogin):
        url = self._url(resource="/pyrasgo-login", api_version=1)
        response = requests.post(url, json=payload.__dict__)
        status_code = response.status_code
        if status_code == 200:
            return response.json()
        if status_code == 401 or status_code == 403:
            raise errors.APIError(
                "Username or password are incorrect. Please check your credentials "
                "or use pyrasgo.register() to create a new account."
            )
        elif status_code == 400:
            raise errors.APIError("Credentials Expired. Contact Rasgo Support.")

    def _url(self, resource, api_version=None):
        env = Environment.from_environment()
        if '/' == resource[0]:
            resource = resource[1:]
        protocol = 'http' if env.value == 'localhost' else 'https'
        return f"{protocol}://{env.value}/{'' if api_version is None else f'v{api_version}/'}{resource}"



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field


class InsightCreate(BaseModel):
    name: Optional[str]
    dataset_id: int = Field(alias='datasetId')
    transform_id: int = Field(alias='transformId')
    transform_arguments: Dict[str, Any] = Field(alias='transformArguments')
    dw_table_id: int = Field(alias='dwTableId')

    class Config:
        allow_population_by_field_name = True


class Insight(InsightCreate):
    id: int



"""
Snowflake DataWarehouse Class
"""
from typing import Dict, List, Optional, Tuple
from typing_extensions import Literal

import pandas as pd

from pyrasgo.api.session import Session
from pyrasgo.storage.dataframe import utils as dfutils
from pyrasgo import errors
from pyrasgo.imports import sf_connector, write_pandas
from pyrasgo.storage.datawarehouse import utils


class SnowflakeDataWarehouse(Session):
    """
    Snowflake DataWarehouse Class
    """

    def __init__(self):
        if not sf_connector:
            raise errors.PackageDependencyWarning(
                "Missing a required python package to run Snowflake. "
                "Please download the Snowflake package by running: "
                "pip install pyrasgo[snowflake]"
            )

        _creds = self.profile.get("dw_creds")
        _organization = self.profile.get("organization")
        if not _creds:
            raise errors.DWCredentialsWarning("Your user is missing credentials, please contact Rasgo.")

        self._connection: sf_connector.SnowflakeConnection = None
        self._credentials: dict = {
            "user": _creds.get("user"),
            "password": _creds.get("password"),
            "role": _creds.get("role", _organization.get("role")),
            "account": _creds.get("account", _organization.get("account")),
            "database": _creds.get("database", _organization.get("database")),
            "schema": _creds.get("schema", _organization.get("schema")),
            "warehouse": _creds.get("warehouse", _organization.get("warehouse")),
            "application": "rasgo",
            "session_parameters": {"QUERY_TAG": "rasgo_python_sdk"},
        }
        self.default_database = _organization.get("database")
        self.default_schema = _organization.get("schema")

    @property
    def connection(self) -> 'sf_connector.SnowflakeConnection':
        """
        Returns a connection to Snowflake
        """
        if not self._connection:
            self._connection = sf_connector.connect(**self._credentials)
        return self._connection

    def execute_query(
        self,
        query: str,
        params: Optional[dict] = None,
        **kwargs,  # Added to accept & ignore args from other dws
    ):
        """
        Execute a query against this DataWarehouse and return results as a list.

        Args:
            query: SQL string to be executed

        Returns:
            List of Tuples
        """
        return self.connection.cursor().execute(query, params)

    def get_schema(
        self,
        fqtn_or_sql: str,
    ) -> Dict[str, str]:
        """
        Return the schema of a table, view, or select statement

        Params:
        `fqtn_or_sql`: str:
            Either a Fully-qualified table name (database.schema.table)
            or a SQL select statement that will create a view.
        """
        # Check for SQL
        if utils.is_select_statement(fqtn_or_sql):
            query_response = self.connection.cursor().describe(fqtn_or_sql)
            return {row[0]: utils.convert_to_type(row[1], row[4], row[5]) for row in query_response}
        # Otherwise assume fqtn:
        query_response = self.query_into_dict(f"DESC TABLE {fqtn_or_sql}")
        return {row['name']: row['type'] for row in query_response}

    def query_into_dict(
        self,
        query: str,
        params: Optional[dict] = None,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> pd.DataFrame:
        """
        Execute a query against this DataWarehouse and return results as a dictionary.

        Args:
            query: SQL string to be executed

        Returns:
            List of Dicts
        """
        return self.connection.cursor(sf_connector.DictCursor).execute(query, params).fetchall()

    def query_into_dataframe(
        self,
        query: str,
        params: Optional[dict] = None,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> pd.DataFrame:
        """
        Execute a query against this DataWarehouse and return results as a pandas DataFrame.

        Args:
            query: SQL select statement to be executed

        Returns:
            pandas DataFrame
        """
        cur = self.connection.cursor()
        cur.execute(query, params)
        return cur.fetch_pandas_all()

    def list_table_columns(
        self,
        fqtn: str,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> List[Tuple[str, str]]:
        """
        Return a list of columns in a table or view

        Args:
            fqtn: Fully-qualified table name (database.schema.table)

        Returns:
            List of Tuples
        """
        desc_sql = f"DESC TABLE {fqtn}"
        query_response = self.query_into_dict(desc_sql)
        return [(x["name"], x["type"]) for x in query_response]

    def make_select_statement(
        self,
        table_metadata: Dict[str, str],
        filters: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
        limit: Optional[int] = None,
        **kwargs,  # Added to accept & ignore args from other dws
    ) -> str:
        """
        Constructs & returns a SQL select statement

        Args:
            table_metadata: Dict must include fqtn, or database + schema + table
            filters: List of SQL WHERE filters strings to filter on in query
            order_by: List of columns to order by in query
            columns: List of columns to return in the query
            limit: Limit the number of rows returned in the query

        Returns:
            string
        """
        # Get table fqtn
        fqtn = table_metadata.get("fqtn")
        if not fqtn:
            fqtn = "{database}.{schema}.{table}".format(**table_metadata)

        # Create Initial select statement
        columns = ", ".join(columns) if columns else "*"
        query = f"SELECT {columns} FROM {fqtn}"

        # Add filters, orders, and limit if supplied
        if filters:
            query = f"{query} WHERE {' AND '.join(filters)}"
        if order_by:
            query = f"{query} ORDER BY {', '.join(order_by)}"
        if isinstance(limit, int):
            query = f"{query} LIMIT {limit}"
        return query.strip()

    def write_dataframe_to_table(
        self,
        df: pd.DataFrame,
        table_name: str,
        method: Literal["append", "overwrite", "fail"] = "fail",
        **kwargs,  # Added to accept & ignore args from other dws
    ):
        """
        Uploads a pandas DataFrame into this DataWarehouse as a table

        Args:
            df: pandas DataFrame to upload
            table_name: name for the target table in the DataWarehouse
            method: Literal:
                append = if table_name already exists, append to it
                overwrite = if table_name already exists, overwrite it
                fail = if table_name already exists, fail

        Returns:
            None
        """
        # Create the table
        dfutils.cleanse_sql_dataframe(df)
        make_table_query = dfutils.generate_ddl(df=df, table_name=table_name, method=method)
        self.execute_query(make_table_query)
        # Upload the df into it
        write_pandas(
            conn=self.connection,
            df=df,
            table_name=table_name,
            quote_identifiers=False,
        )



"""
sql rendering functions
"""
__all__ = ['operations_as_cte', 'offline_operations_as_cte']
import re
from typing import List, Dict, Any

from rasgotransforms.render import RasgoEnvironment
import pandas as pd

from pyrasgo.schemas.dw_operation import Operation, OperationCreate
from pyrasgo.schemas.transform import Transform
from pyrasgo.primitives.dataset import Dataset
from pyrasgo.utils import naming
from pyrasgo.constants import SOURCE_TABLE_ARG_NAME


def operation_as_subquery(source_code: str, source_table: str, arguments: Dict[str, Any], running_sql: str = '') -> str:
    from pyrasgo.api import Read

    def run_query(query) -> pd.DataFrame:
        query = f'{running_sql} {query}'
        result = Read().data_warehouse.query_into_dataframe(query)
        return result

    def get_columns(table: str) -> Dict[str, str]:
        if running_sql:
            return Read().data_warehouse.get_schema(f"{running_sql} SELECT * FROM {table}")
        else:
            return Read().data_warehouse.get_schema(table)

    env = RasgoEnvironment(run_query=run_query)
    return env.render(
        source_code=source_code,
        source_table=source_table,
        arguments=arguments,
        override_globals={'get_columns': get_columns},
    )


def offline_operations_as_cte(operations: List[OperationCreate], transforms: List[Transform]) -> str:
    if operations:
        transforms = {t.id: t for t in transforms}
        sql = ''
        sub_queries = []
        for i in range(len(operations)):
            arguments = replace_args(operations[i].operation_args.copy(), sub_queries)
            if transforms[operations[i].transform_id].name == 'apply':
                source_code = arguments['sql']
            else:
                source_code = transforms[operations[i].transform_id].sourceCode
            operation_sql = operation_as_subquery(
                source_code=source_code,
                source_table=arguments[SOURCE_TABLE_ARG_NAME],
                running_sql=sql,
                arguments=arguments,
            )
            sub_queries.append(operations[i].table_name)
            sql += 'WITH ' if i == 0 else ', '
            sql += f"{operations[i].table_name} as (\n{operation_sql}\n)"
        last_op = operations[-1].table_name
        sql += f' SELECT * FROM {last_op}'
        return sql


def replace_args(arguments, sub_queries):
    """
    Traverses all args recursively and replaces instances of fqtns with table names if the tables are created in the cte
    """

    def replace_string(s):
        if isinstance(s, Dataset):
            s = s.fqtn
        if isinstance(s, str) and s.split('.')[-1] in sub_queries:
            return s.split('.')[-1]
        else:
            return s

    def replace_list(l):
        l = l.copy()
        for i in range(len(l)):
            if isinstance(l[i], dict):
                l[i] = replace_dict(l[i])
            elif isinstance(l[i], list):
                l[i] = replace_list(l[i])
            else:
                l[i] = replace_string(l[i])
        return l

    def replace_dict(d):
        d = d.copy()
        for k in d.keys():
            if isinstance(d[k], dict):
                d[k] = replace_dict(d[k])
            if isinstance(d[k], list):
                d[k] = replace_list(d[k])
            else:
                d[k] = replace_string(d[k])
        return d

    return replace_dict(arguments.copy())


def operations_as_cte(
    operations: List[Operation],
) -> str:
    """
    Returns a nested CTE statement to render this op set as a CTE
    """
    # Handle single transform chains, we already have the SQL
    if len(operations) == 1:
        return operations[0].operation_sql

    # Handle multi-transform chains
    operation_list = []
    # Need to replace old FQTNs with CTE aliases
    fqtn_mapping = {}
    for operation in operations:
        # create new aliases to replace old FQTNs with
        alias = operation.dw_table.table_name or naming.random_table_name()
        fqtn_mapping[operation.dw_table.fqtn] = alias

        op_sql = operation.operation_sql

        # replace all instances of generated FQTNs with CTE aliases
        for fqtn, alias in fqtn_mapping.items():
            op_sql = op_sql.replace(fqtn, alias)

        # if final op, we're done. join ops and leave last one alone
        # Final op itself might be a CTE, remove the WITH and slap it on the end of this chain
        if operation == operations[-1]:
            return 'WITH {}{}'.format(', \n'.join(operation_list), collapse_cte(op_sql))

        operation_list.append(f'{alias} AS (\n{op_sql}\n) ')


def collapse_cte(sql: str) -> str:
    """
    Returns a collapsed CTE if the sql itself is already a CTE (starts with "with")
    """
    return re.sub(r'^(WITH)\s', ', ', sql, 1, flags=re.IGNORECASE)



from enum import Enum
from typing import Optional, List, Any, Dict

from pydantic import BaseModel, Field


class AcceleratorOperationType(str, Enum):
    """
    The type of operation to create.

    OPERATION: Creates a canvas node with the given inputs
    INSIGHT: Creates an insight chart with the given inputs
    """

    OPERATION = 'OPERATION'
    INSIGHT = 'INSIGHT'


class AcceleratorOperationCreate(BaseModel):
    """
    An operation to create for an accelerator

    Attributes:
        name:
            The name of the operation. Must be a variable-safe name. Can be used
            to reference this operation in subsequent steps
        description:
            Optional description of this operation step
        transform_id:
            id of the transform to apply in creating this operation
        transform_arguments:
            Arguments to apply to the transform.
            Arguments *must* be
                - A Jinja template, parsable with Jinja syntax
                - Valid JSON. After Jinja parsing, will be loaded as python objects
                    using `json.loads()` and passed as transform arguments
            Special Argument:
                json_string_arguments:
                    The string passed in this argument will be Jinja and JSON parsed
                    like the others, then added to the transform_arguments. This allows
                    for building complex objects using Jinja (e.g. passing multiple
                    transform arguments based on one user argument)
        operation_type:
            Enum AcceleratorOperationType
            Operations of type OPERATION will be created as dataset operations
            Operations of type INSIGHT will  be created as Insight charts on the dataset
    """

    name: str
    description: Optional[str]
    transform_id: int = Field(alias='transformId')
    transform_arguments: Dict[str, Any] = Field(alias='transformArguments')
    operation_type: AcceleratorOperationType = AcceleratorOperationType.OPERATION

    class Config:
        allow_population_by_field_name = True
        use_enum_values = True


class AcceleratorArgumentCreate(BaseModel):
    """
    An argument to require from users that will be passed to Accelerator operations

    Attributes:
        name:
            The name of the argument. Must be variable-safe and will be passed to operations
        description:
            Description of the argument and how to use it
        argument_type:
            The argument type, consistent with Transforms. Used by Rasgo to determine
            what input type to provide for UI forms and other special cases
            Special Type:
                dataset:
                    Arguments of type `dataset` will expect a dataset ID and will be
                    translated to FTQNs during parsing. These wll also be marked as
                    dependencies for the dataset
        optional:
            Is the argument required
        context:
            Generic data for use in displaying argument to users.
            e.g. context about which table a column argument should come from.
    """

    name: str
    description: Optional[str]
    argument_type: str = Field(alias='type')
    optional: Optional[bool] = Field(alias='optional')
    context: Optional[Dict[str, Any]]

    class Config:
        allow_population_by_field_name = True


class AcceleratorCreate(BaseModel):
    """
    The required fields for creating an Accelerator


    Attributes:
        name:
            The name of the Accelerator
        description:
            A short description of how to use the Accelerator
        arguments:
            Valid list of AcceleratorArguments
        operations:
            Valid list of AcceleratorOperations
    """

    name: str
    description: Optional[str]
    arguments: Optional[List[AcceleratorArgumentCreate]]
    operations: List[AcceleratorOperationCreate]


class Accelerator(AcceleratorCreate):
    """
    A full Accelerator definition, complete with required arguments and definitions
    for what operations an Accelerator will create
    """

    id: int


class AcceleratorApply(BaseModel):
    """
    Represents the arguments that get passed so some given Accelerator to create a Dataset

    Attributes:
        name:
            The name of the Dataset that will be created by the Accelerator
        arguments:
            A dictionary of arguments to pass to the Accelerator.
    """

    name: Optional[str]
    arguments: Dict[str, Any]



"""
dbt functions
"""
import copy
import logging
import os
from pathlib import Path
import re
import string
from typing import Any, Dict, List, Union

import yaml

from pyrasgo.primitives import Dataset, DbtModel, DbtSource, DbtSourceParent

__all__ = [
    "dataset_to_model",
    "dataset_to_source",
    "save_model_file",
    "write_dbt_files",
]


logger = logging.getLogger("dbt")

DBT_MODEL_CONFIG_TEMPLATE = """
  config(
    {config_args}
  )
"""

DBT_PROJECT_TEMPLATE = {
    "name": "",
    "version": "1.0.0",
    "config-version": 2,
    "profile": "default",
    "model-paths": ["models"],
    "analysis-paths": ["analyses"],
    "test-paths": ["tests"],
    "seed-paths": ["seeds"],
    "macro-paths": ["macros"],
    "snapshot-paths": ["snapshots"],
    "target-path": "target",
    "log-path": "logs",
    "packages-install-path": "dbt_packages",
    "clean-targets": ["target", "dbt_packages"],
    "models": None,
}


def check_project_name(project_name: str):
    """
    Checks a project name for dbt compliance
    """
    allowed = set(string.ascii_lowercase + "_")
    if any(char for char in project_name.lower() if char not in allowed):
        logger.warning(
            "per dbt: Project names should contain only lowercase characters "
            "and underscores. A good package name should reflect your organization's "
            "name or the intended use of these models"
        )
    return project_name.lower()


def dataset_to_model(dataset: Dataset) -> DbtModel:
    """
    Primitive conversion: accepts a Rasgo Dataset and returns a DbtModel
    """
    from pyrasgo.api import Get

    metrics = Get().metrics(dataset.id)
    database, schema, table = dataset.fqtn.split(".")
    return DbtModel(
        name=table,
        fqtn=dataset.fqtn,
        sql_definition=dataset.sql,
        columns_schema=dataset.schema,
        config_args={"database": database, "schema": schema, "+materialized": dataset.table_type},
        metrics=metrics,
    )


def dataset_to_source(dataset: Dataset) -> DbtSource:
    """
    Primitive conversion: accepts a Rasgo Dataset and returns a DbtSource
    """
    database, schema, table = dataset.fqtn.split(".")
    return DbtSource(
        name=table,
        fqtn=dataset.fqtn,
        parent=DbtSourceParent(
            name=f"{database}_{schema}",
            config_args={"database": database, "schema": schema},
        ),
        columns_schema=dataset.schema,
    )


def determine_sources_and_refs(
    models: List[DbtModel],
    known_sources: List[DbtSource],
):
    """
    Finds fqtns in all models and splits them into two lists:
    refs: tables created by one of these models
    sources: tables created prior to / outside of these models
    """
    fqtn_set = set()
    source_set = {s.fqtn for s in known_sources}
    ref_dict = {m.fqtn: m.name for m in models}
    for model in models:
        fqtn_set.update(extract_fqtns(model.sql_definition))
    source_set.update({t for t in fqtn_set if t not in ref_dict.keys()})
    source_dict = {f"{d}.{s}.{n}": {"parent": f"{d}_{s}", "name": n} for d, s, n in (t.split(".") for t in source_set)}
    return source_dict, ref_dict


def extract_fqtns(sql_definition: str) -> List[str]:
    """
    Returns all fully qualified table names in a sql string
    """
    fqtn_list = re.findall(r"[^\w]+\.[^\w]+\.[^\w]+", sql_definition)
    return fqtn_list


def prepare_dbt_path(project_directory: Union[os.PathLike, str]) -> os.PathLike:
    """
    Checks for a specified filepath and creates one if it doesn't exist
    """
    project_directory = Path(project_directory)
    for dir_name in ("analyses", "dbt_packages", "logs", "macros", "models", "seeds", "target", "tests"):
        (project_directory / dir_name).mkdir(exist_ok=True, parents=True)
    return project_directory


def replace_sources_and_refs(
    sql_definition: str,
    model_name: str,
    ref_tables: Dict[str, str] = None,
    source_tables: Dict[str, Dict[str, str]] = None,
):
    """
    Combs a sql statement for fqtns and replaces them with dbt ref or source functions
    """
    if ref_tables:
        for fqtn, alias in ref_tables.items():
            if alias != model_name:
                sql_definition = sql_definition.replace(fqtn, "{} ref('{}') {}".format("{{", alias, "}}"))
    if source_tables:
        for fqtn, table_parts in source_tables.items():
            sql_definition = sql_definition.replace(
                fqtn, "{} source('{}', '{}') {}".format("{{", table_parts.get("parent"), table_parts.get("name"), "}}")
            )
    return sql_definition


def save_project_file(
    project_name: str,
    project_directory: Union[os.PathLike, str],
    overwrite: bool = False,
    model_args: Dict[str, Any] = None,
) -> Path:
    """
    Writes a yaml definition to a dbt project file
    """
    filepath = Path(project_directory) / "dbt_project.yml"
    if not os.path.exists(filepath) or overwrite:
        yml_definition = copy.deepcopy(DBT_PROJECT_TEMPLATE)
        yml_definition["name"] = project_name
        model_config = {project_name: model_args}
        yml_definition["models"] = model_config
        with open(filepath, "w") as _yaml:
            yaml.dump(data=yml_definition, Dumper=yaml.SafeDumper, stream=_yaml, sort_keys=False)
    return filepath


def save_model_file(
    sql_definition: str,
    output_directory: Union[os.PathLike, str],
    file_name: str,
    ref_tables: Dict[str, str] = None,
    source_tables: Dict[str, Dict[str, str]] = None,
) -> str:
    """
    Writes a sql script to a dbt model file
    """
    output_directory = Path(output_directory)
    file_name = f"{file_name}.sql" if file_name[-4:] != ".sql" else file_name
    model_name = file_name.replace(".sql", "")
    filepath = output_directory / file_name
    os.makedirs(output_directory, exist_ok=True)
    sql_definition = replace_sources_and_refs(
        sql_definition=sql_definition,
        model_name=model_name,
        ref_tables=ref_tables,
        source_tables=source_tables,
    )
    with open(filepath, "w") as _file:
        _file.write(sql_definition)
    return filepath


def save_schema_file(
    output_directory: Union[os.PathLike, str],
    file_name: str = None,
    models: List[DbtModel] = None,
    sources: List[DbtSource] = None,
) -> str:
    """
    Writes dbt model and source details to a schema yml file
    """
    file_name = file_name or "schema.yml"
    file_name = f"{file_name}.yml" if file_name[-4:] not in (".yml", "yaml") else file_name
    filepath = Path(output_directory) / file_name

    existing_schema = None
    if os.path.exists(filepath):
        with open(filepath, "r") as _file:
            existing_schema = yaml.safe_load(_file)
    schema_definition = existing_schema or {"version": 2, "sources": [], "models": [], "metrics": []}

    # Add sources to schema
    if sources and not schema_definition.get("sources"):
        schema_definition["sources"] = []
    for source in sources:
        parent_found = False
        source_found = False
        columns_list = [{"name": row[0]} for row in source.columns_schema]
        for p in schema_definition["sources"]:
            if p.get("name") == source.parent.name:
                parent_found = True
                for t in p.get("tables"):
                    if t.get("name") == source.name:
                        t["columns"] = columns_list
                        if source.config_args:
                            for k, v in source.config_args.items():
                                t[k] = v
                        source_found = True
        if not parent_found:
            schema_definition["sources"].append({"name": source.parent.name})
        if not source_found:
            source_dict = {"name": source.name, "columns": columns_list}
            if source.config_args:
                source_dict.update(source.config_args)
            for p in schema_definition["sources"]:
                if p.get("name") == source.parent.name:
                    if source.parent.config_args:
                        for k, v in source.parent.config_args.items():
                            p[k] = v
                    if not p.get("tables"):
                        p["tables"] = []
                    p["tables"].append(source_dict)

    # Add models to schema
    if models and not schema_definition.get("models"):
        schema_definition["models"] = []
    for model in models:
        model_found = False
        columns_list = [{"name": row[0]} for row in model.columns_schema]
        for m in schema_definition["models"]:
            if m.get("name") == model.name:
                m["columns"] = columns_list
                if model.config_args:
                    m["config"] = model.config_args
                model_found = True
        if not model_found:
            model_dict = {"name": model.name, "columns": columns_list}
            if model.config_args:
                model_dict.update({"config": model.config_args})
            schema_definition["models"].append(model_dict)
        if model.metrics and not schema_definition.get("metrics"):
            schema_definition["metrics"] = []
        for metric in model.metrics:
            for m in schema_definition["metrics"]:
                if m.get("name") == metric.name:
                    m["label"] = metric.label
                    m["model"] = f"ref('{model.name}')"
                    m["description"] = metric.description
                    m["type"] = metric.aggregation_type
                    m["sql"] = metric.target_expression
                    m["timestamp"] = metric.time_dimension
                    m["time_grains"] = metric.time_grains
                    m["dimensions"] = metric.dimensions
                    m["filters"] = metric.filters or []
                    m["meta"] = metric.meta or {}
                    break
            else:
                metric_dict = {
                    "name": metric.name,
                    "label": metric.label,
                    "model": f"ref('{model.name}')",
                    "description": metric.description,
                    "type": metric.aggregation_type,
                    "sql": metric.target_expression,
                    "timestamp": metric.time_dimension,
                    "time_grains": metric.time_grains,
                    "dimensions": metric.dimensions,
                    "filters": metric.filters or [],
                    "meta": metric.meta or {},
                }
                schema_definition["metrics"].append(metric_dict)
    with open(filepath, "w") as _file:
        yaml.dump(data=schema_definition, Dumper=yaml.SafeDumper, stream=_file, sort_keys=False)
    return filepath


def write_dbt_files(
    project_name: str,
    project_directory: Union[os.PathLike, str],
    models_directory: Union[os.PathLike, str],
    models: List[DbtModel],
    sources: List[DbtSource],
    model_args: Dict[str, Any] = None,
    verbose: bool = False,
) -> str:
    """
    Saves a dbt_project.yml and model.sql files to a directory
    """
    check_project_name(project_name)
    project_directory = Path(project_directory)
    models_directory = Path(models_directory)
    prepare_dbt_path(project_directory)
    save_project_file(
        project_name=project_name,
        project_directory=project_directory,
        model_args=model_args,
    )
    source_tables, ref_tables = determine_sources_and_refs(models, sources)
    i = 0
    for model in models:
        # Write a sql and yml file for each model
        if verbose:
            i += 1
            print(f"writing model {i} of {len(models)}")
        save_model_file(
            sql_definition=model.sql_definition,
            output_directory=models_directory,
            file_name=f"{model.name}.sql",
            ref_tables=ref_tables,
            source_tables=source_tables,
        )
        save_schema_file(
            output_directory=models_directory,
            file_name=f"{model.name}.yml",
            models=[model],
            sources=[],
        )
    # Write a single schema file for all sources
    save_schema_file(
        output_directory=models_directory,
        file_name="sources.yml",
        models=[],
        sources=sources,
    )
    return project_directory



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Optional

from pydantic import BaseModel


class StatusTracking(BaseModel):
    tracking_uuid: str
    message: Optional[str] = None
    status: str



"""
API Delete functions
"""
from pyrasgo import errors


class Delete:
    """
    API Delete Class
    """

    def __init__(self):
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    def accelerator(self, accelerator_id: int) -> str:
        """
        Delete an Accelerator in Rasgo by id
        """
        response = self.api._delete(f"/accelerators/{accelerator_id}", api_version=2)
        if response.status_code == 200:
            return f"Accelerator with id '{accelerator_id}' successfully deleted"
        return f"Problem deleting Dataset {accelerator_id}."

    def dataset(self, dataset_id: int) -> str:
        """
        Delete a Dataset in Rasgo
        """
        response = self.api._delete(f"/datasets/{dataset_id}", api_version=2)
        if response.status_code == 200:
            return f"Dataset with id '{dataset_id}' successfully deleted"
        return f"Problem deleting Dataset {dataset_id}."

    def metric(self, metric_id: int) -> str:
        """Delete a Rasgo Metric by ID"""
        response = self.api._delete(f"/metric/{metric_id}", api_version=2)
        if response.status_code == 200:
            return f"Metric with id '{metric_id}' successfully deleted"
        elif response.status_code == 404:
            raise errors.RasgoResourceException(
                f"Could not find metric with id '{metric_id}' or user does not have access to delete"
            )
        return f"Could not delete metric '{metric_id}'"

    def transform(self, transform_id: int) -> str:
        """
        Delete a Rasgo User Defined Transform
        """
        # NOTE: We print out error msgs on the API side
        # in the function self.api._raise_internal_api_error_if_any(response)
        # so no need to print out logic like above
        response = self.api._delete(f"/transform/{transform_id}", api_version=1)
        if response.status_code == 200:
            return f"Transform with id '{transform_id}' successfully deleted"
        if response.status_code == 403:
            raise errors.RasgoResourceException(
                f"User does not have access to delete Transform with id '{transform_id}'"
            )
        return f"Problem deleting Transform {transform_id}."



from pydantic import BaseModel, Field


class UserRegistration(BaseModel):
    email: str
    password: str


class UserLogin(BaseModel):
    email: str
    password: str


class User(BaseModel):
    id: int
    username: str
    email: str
    api_key: str = Field(alias='key')
    organization_id: int = Field(alias='organizationId')

    class Config:
        allow_population_by_field_name = True



from pyrasgo.api import Create, Delete, Get, Publish, Read, Update


class Rasgo:
    """
    Base connection object to handle interactions with the Rasgo API.
    """

    def __init__(self):
        self._experiment_id = None
        self.create = Create()
        self.delete = Delete()
        self.get = Get()
        self.publish = Publish()
        self.read = Read()
        self.update = Update()

    def _pyrasgo_api_key(self):
        from pyrasgo.config import get_session_api_key

        return get_session_api_key()



import time

from pyrasgo import errors, schemas


def poll_operation_set_async_status(
    task_request: schemas.OperationSetAsyncTask,
    delay_start=1,
    delay_max=1.5,
    backoff_multiplier=1.5,
    verbose=False,
) -> int:
    """
    Given some task request, this function will poll the API until the task hits a terminal status.

    If operation set created successfully with no errors, return operation set id
    """
    from pyrasgo.api import Get

    get = Get()

    operations_expected = len(task_request.request.operations)
    operations_created = 0
    delay = delay_start
    while True:
        time.sleep(delay)
        task_request = get._operation_set_async_status(task_request.id)

        # Get the status of the last event
        last_event_status = task_request.events[-1].event_type

        # Let the user know if this started
        if last_event_status == 'STARTED':
            if verbose:
                print(f'Operation set task {task_request.id} has started')

        # Is this progress? Keep track.
        if last_event_status == 'PROGRESSED':
            operations_created = task_request.events[-1].message['complete']
            if verbose:
                print(f'Operation set creation: {operations_created} of {operations_expected} operations complete')

        # Is this a failure? Throw an exception
        if last_event_status == 'FAILED':
            raise errors.APIError(
                f'Operation set creation failure: {task_request.events[-1].message}; '
                f'Progress: {operations_created} / {operations_expected}'
            )

        # Is this a success? Get out of this loop
        if last_event_status == 'FINISHED':
            operation_set_id = task_request.events[-1].message['operation_set_id']
            return operation_set_id

        # Still waiting? Grab a Snickers
        if delay < delay_max:
            delay = min(delay * backoff_multiplier, delay_max)
        if verbose:
            print('Poll waiting...')



from typing import Optional


def is_select_statement(
    sql: str,
) -> bool:
    """
    Determine if a string is a valid SQL select statement and return a boolean
    """
    if sql.lower().startswith("with") and "select" in sql.lower():
        return True
    if sql.lower().startswith("select"):
        return True
    return False


def convert_to_type(type_code: int, precision: int, scale: int) -> str:
    """
    Accepts values from a Snowflake ResultMetadata record
    and returns a Snowflake data type

    https://docs.snowflake.com/en/user-guide/python-connector-api.html#type-codes
    ResultMetadata(name='', type_code=, display_size=, internal_size=, precision=, scale=, is_nullable=)
    """
    type_code_map = {
        0: "FIXED",  # NUMBER/INT
        1: "REAL",  # REAL
        2: "TEXT",  # VARCHAR/STRING
        3: "DATE",  # DATE
        4: "TIMESTAMP",  # TIMESTAMP
        5: "VARIANT",  # VARIANT
        6: "TIMESTAMP_LTZ",  # TIMESTAMP_LTZ
        7: "TIMESTAMP_TZ",  # TIMESTAMP_TZ
        8: "TIMESTAMP_NTZ",  # TIMESTAMP_TZ
        9: "OBJECT",  # OBJECT
        10: "ARRAY",  # ARRAY
        11: "BINARY",  # BINARY
        12: "TIME",  # TIME
        13: "BOOLEAN",  # BOOLEAN
    }
    data_type = type_code_map[type_code]
    if data_type == 'REAL':
        return 'FLOAT'
    if data_type == 'FIXED':
        return f'NUMBER({precision or 0},{scale or 0})'
    if 'TIMESTAMP' in data_type:
        return f'{data_type}({scale or 0})'
    return data_type



# Value of dataset.status if dataset is published
DS_PUBLISHED_STATUS = 'published'

# Name of required source table arg we auto populate
SOURCE_TABLE_ARG_NAME = 'source_table'

# Transform Arg Types which expect DS inputs
# should contain Datasets Objs
TABLE_ARG_TYPE = 'table'
TABLE_LIST_ARG_TYPE = 'table_list'



"""
API Create Functions
"""
import time
from typing import Any, Dict, List, Optional, Union
from urllib import parse

try:
    from typing import Literal  # Only available on 3.8+ directly
except ImportError:
    from typing_extensions import Literal

import yaml

from pyrasgo.config import MAX_POLL_ATTEMPTS, POLL_RETRY_RATE
from pyrasgo.utils import polling
from pyrasgo import primitives, schemas, errors
from pyrasgo.schemas.dataset import DatasetSourceType
from pyrasgo.schemas.offline import OfflineDataset


class Create:
    """
    API Create Class
    """

    def __init__(self):
        from pyrasgo.api import Get
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self._get = Get()

    def accelerator(self, *, accelerator_create: schemas.AcceleratorCreate) -> schemas.Accelerator:
        """
        Create a new accelerator Definition
        """
        # Create Accelerator in API and return the new object
        resp = self.api._post("/accelerators", _json=accelerator_create.dict(), api_version=2).json()
        return schemas.Accelerator(**resp)

    def accelerator_from_yaml(self, *, yaml_string: str) -> schemas.Accelerator:
        """
        Create a new accelerator Definition from a yml string
        Parses yml document to correct format that Rasgo expects

        Example format:
            name: Accelerator Name
            description: Accelerator Description
            arguments:
                base_dataset:
                    description: the base dataset
                    type: dataset
                column_name:
                    description: the column to drop
                    type: column
                column_name_two:
                    description: the column to drop
                    type: column
            operations:
                operation_1_name:
                    description: Drop a column
                    transform_name: drop_columns
                    transform_arguments:
                        source_table: '{{dataset_id}}'
                        json_string_arguments: '{"exclude_cols": ["{{column_name}}"]}'
                operation_2_name:
                    description: Drops another column
                    transform_name: drop_columns
                    transform_arguments:
                        source_table: '{{operation_1_name}}'
                        json_string_arguments: '{"exclude_cols": ["{{column_name_two}}"]}'
        """
        return self.accelerator(accelerator_create=self._build_accelerator(yaml_string))

    def dataset_from_accelerator(self, id: int, arguments: Dict[str, Any], name: str) -> None:
        """
        Applies the given set of arguments to an Accelerator to generate
        a new DRAFT Dataset in the Rasgo, with the given name

        Args:
            id: Id of Accelerator to apply
            arguments: Arguments of the Accelerator
            name: Name of the created dataset
        """
        if not name:
            raise errors.RasgoRuleViolation("Please supply a name for the Dataset to create with an Accelerator")

        # Create a Draft Dataset in the Rasgo from the Accelerator
        apply_request = schemas.AcceleratorApply(name=name, arguments=arguments)
        resp = self.api._post(f"/accelerators/{id}/apply", _json=apply_request.dict(), api_version=2)

        # Print message telling users Dataset creation in progress,
        # give the URL to access it, and tell them how to get the
        # PyRasgo code of the created Accelerator Dataset
        api_dataset = schemas.Dataset(**resp.json())
        draft_dataset = primitives.Dataset(api_dataset=api_dataset)
        print(
            f"Draft dataset named {name!r} with id {draft_dataset.id} is being created in Rasgo.\n"
            f"View it's creation progress at {draft_dataset.profile()}\n\n"
            f"After the dataset is finished being built, get the PyRasgo code to re-create it using\n"
            f"    ds = rasgo.get.dataset({draft_dataset.id})\n"
            f"    print(ds.generate_py())"
        )

    def dataset_from_dict(
        self,
        dataset_dict: dict,
    ) -> primitives.Dataset:
        """
        Creates a new version of a dataset based on a Rasgo-compliant dict

        Args:
            dataset_dict: a Rasgo-compliant json dict (converted from a yaml file)

        To create a dataset from a yaml file:
            with open(your_file_path, "r") as yaml_file:
                rasgo.create.dataset_from_json(yaml.safe_load(yaml_file))
        """
        try:
            dataset_v2 = OfflineDataset(**dataset_dict)
        except Exception as ex:
            raise errors.RasgoSchemaException("dataset_dict violates Rasgo-compliant format") from ex
        response = self.api._post("/datasets/from-offline-version", dataset_v2.dict(), api_version=2).json()
        api_dataset = schemas.Dataset(**response)
        draft_dataset = primitives.Dataset(api_dataset=api_dataset)
        print(
            f"Draft dataset {draft_dataset.name} (resource_key={draft_dataset.resource_key}, id={draft_dataset.id}, version={draft_dataset.version}) is being created in Rasgo.\n"
            f"View it's creation progress at {draft_dataset.profile()}\n\n"
            f"After the dataset is built, you can access it in pyrasgo using:\n"
            f"    ds = rasgo.get.dataset({draft_dataset.id})\n"
        )
        return draft_dataset

    def dataset_snapshot(self, id: int) -> str:
        """
        Takes a snapshot of a dataset

        Args:
            id: Id of dataset to snapshot
        """
        self.api._post("/datasets/snapshots", _json={"datasetId": id}, api_version=2)
        return (
            "Sent request to build dataset snapshot. "
            f"Run rasgo.get.dataset({id}).snapshots to see available snapshots."
        )

    def metric(
        self,
        name: str,
        dataset_id: int,
        aggregation_type: str,
        target_expression: str,
        time_grains: List[Literal["HOUR", "DAY", "WEEK", "MONTH", "QUARTER", "YEAR"]],
        time_dimension: str,
        dimensions: List[str],
        filters: Optional[List[schemas.Filter]] = None,
        meta: Optional[Dict[str, str]] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> schemas.Metric:
        """
        Creates a metric on a dataset
        """
        time_grain_objs = [schemas.metric.TimeGrain[grain] for grain in time_grains]
        metric = schemas.MetricCreate(
            ds_dataset_id=dataset_id,
            name=name,
            aggregation_type=aggregation_type,
            target_expression=target_expression,
            time_grains=time_grain_objs,
            time_dimension=time_dimension,
            dimensions=dimensions,
            filters=filters,
            meta=meta,
            label=label,
            description=description,
        )
        try:
            response = self.api._post(
                f"/metric", metric.dict(exclude_unset=True, exclude_none=True), api_version=2
            ).json()
            return schemas.Metric(**response)
        except Exception as err:
            raise errors.RasgoResourceException(f"Could not create metric on dataset '{dataset_id}'.") from err

    def transform(
        self,
        *,
        name: str,
        source_code: str,
        type: Optional[str] = None,  # TODO: name this something other than a Python reserved word
        arguments: Optional[List[dict]] = None,
        description: Optional[str] = None,
        tags: Optional[Union[List[str], str]] = None,
        context: Optional[Dict[str, Any]] = None,
        dw_type: Optional[Literal["SNOWFLAKE", "BIGQUERY", "UNSET"]] = None,
    ) -> schemas.Transform:
        """
        Create and return a new Transform in Rasgo
        Args:
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform
            tags: List of tags, or a tag (string), to set on this dataset
            context: Object used to add context to transforms for client use
            dw_type: DataWarehouse provider: SNOWFLAKE, BIGQUERY or UNSET
                     if not provided, will be set to your current DataWarehouse

        Returns:
            Created Transform obj
        """
        arguments = arguments if arguments else []

        # Init tag array to be list of strings
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        transform = schemas.TransformCreate(
            name=name,
            type=type,
            sourceCode=source_code,
            description=description,
            tags=tags,
            context=context,
            dw_type=dw_type.upper() if dw_type else None,
        )
        transform.arguments = [schemas.TransformArgumentCreate(**x) for x in arguments]
        response = self.api._post("/transform", transform.dict(), api_version=1).json()
        return schemas.Transform(**response)

    # ----------------------------------
    #  Internal/Private Create Calls
    # ----------------------------------

    def _build_accelerator(self, yaml_string: str):
        available_transforms = self._get.transforms()

        # Create Accelerator object
        yaml_dict = yaml.safe_load(yaml_string)
        yaml_dict['arguments'] = [{**{'name': k}, **v} for k, v in yaml_dict['arguments'].items()]
        yaml_dict['operations'] = [{**{'name': k}, **v} for k, v in yaml_dict['operations'].items()]

        # replace transform names with transform Ids
        for i, operation in enumerate(yaml_dict['operations']):
            transform_name = operation.pop('transform_name' if 'transform_name' in operation else 'transformName', None)
            if transform_name:
                transform_available = [x for x in available_transforms if x.name == transform_name]
                if transform_available:
                    operation['transform_id'] = transform_available[0].id
                    yaml_dict['operations'][i] = operation
                else:
                    raise errors.APIError(f'Transform {transform_name} does not exist or is not available')

        return schemas.AcceleratorCreate(**yaml_dict)

    def _dataset(
        self,
        *,
        name: str,
        source_type: DatasetSourceType,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        status: Optional[str] = 'published',
        dw_table_id: Optional[int] = None,
        dw_operation_set_id: Optional[int] = None,
        fqtn: Optional[str] = None,
        attributes: Optional[dict] = None,
        tags: Optional[List[str]] = None,
        publish_ds_table_type: Optional[str] = "VIEW",
        publish_ds_table_name: Optional[str] = None,
        generate_stats: bool = True,
        timeout: Optional[int] = None,
    ) -> schemas.Dataset:
        """
        Create a Dataset in Rasgo.

        Args:
            name: Name of the dataset
            source_type: Type of Dataset we're publishing (Rasgo vs. Snowflake. vs DataFrame)
            description: Description of the dataset
            status: Status of whether this datasets is published or not
            dw_table_id: DW table to associate with this Dataset
            dw_operation_set_id: Id of the Operation Set to associate with this Dataset
            fqtn: Fully qualified table name of the table to register this Dataset as
            attributes: Dictionary containing dataset attributes to be published
            publish_ds_table_type: If publishing an operations set's set DWTable after creating,
                                   select the table type for the published node ('VIEW' or 'TABLE')
            publish_ds_table_name: Optional user supplied table name to set on the
                                   published node/operation
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
        Returns:
            Created Dataset Obj
        """
        if not publish_ds_table_type or publish_ds_table_type.upper() not in ("TABLE", "VIEW"):
            raise errors.ParameterValueError("table_type", ["TABLE", "VIEW"])

        dataset_create = schemas.DatasetCreate(
            name=name,
            resource_key=resource_key,
            description=description,
            status=status,
            dw_table_id=dw_table_id,
            dw_operation_set_id=dw_operation_set_id,
            attributes=attributes,
            tags=tags,
            auto_generate_stats=generate_stats,
            source_type=source_type.value,
        )
        path = "/datasets/async"
        if fqtn:
            path = f"{path}?fqtn={parse.quote(fqtn)}"
        else:
            path = f"{path}?table_type={publish_ds_table_type}"
            if publish_ds_table_name:
                encoded_table_name = parse.quote(publish_ds_table_name)
                path = f"{path}&publish_ds_table_name={encoded_table_name}"
        response = self.api._post(path, dataset_create.dict(), api_version=2).json()
        status_tracking = schemas.StatusTracking(**response)
        for i in range(1, MAX_POLL_ATTEMPTS):
            status_tracking = schemas.StatusTracking(
                **self.api._get(f"/status-tracking/{status_tracking.tracking_uuid}", api_version=2).json()
            )
            if status_tracking.status == "completed":
                return schemas.Dataset(**self.api._get(f"/datasets/{status_tracking.message}", api_version=2).json())
            if status_tracking.status == "failed":
                raise errors.APIError(f"Could not publish dataset: {status_tracking.message}")
            if timeout and (POLL_RETRY_RATE * i) > timeout:
                raise errors.APITimeoutWarning("Timeout reached waiting for dataset creation.")
            time.sleep(POLL_RETRY_RATE)
        raise errors.APIError(
            f"Never received confirmation dataset was published "
            f"after max wait of {(MAX_POLL_ATTEMPTS * POLL_RETRY_RATE) // 60} minutes."
        )

    def _operation_set_non_async(
        self, operations: List[schemas.OperationCreate], dataset_dependency_ids: List[int]
    ) -> schemas.OperationSet:
        """
        Create an operation set in Rasgo with specified operation
        and input dataset dependencies ids, in a  non-async status
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations, dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post("/operation-sets", operation_set_create.dict(), api_version=2).json()
        return schemas.OperationSet(**response)

    def _operation_set_async(
        self, operations: List[schemas.OperationCreate], dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetAsyncTask:
        """
        Create an operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations, dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post("/operation-sets/async", operation_set_create.dict(), api_version=2).json()
        return schemas.OperationSetAsyncTask(**response)

    def _dataset_correlation_stats(self, *, table_id: int, only_if_data_changed: Optional[bool] = True) -> None:
        """
        Trigger stats generation on a dataset
        """

        # Note: dimension_column_id should not be passed, as it has already been set at publish time
        stats_create = schemas.GenerateStat(
            dwTableId=table_id, dimensionColumnId=None, onlyIfDataChanged=only_if_data_changed
        )
        self.api._post("/stats", stats_create.dict(), api_version=2)

    def _operation_render(self, operation: schemas.OperationCreate) -> str:
        """
        Test the rendering of an operation
        """
        response = self.api._post("/operation/render", operation.dict(), api_version=2).json()
        return response

    def _operation_set(
        self,
        operations: List[schemas.OperationCreate],
        dataset_dependency_ids: List[int],
        async_compute: bool = True,
        async_verbose: bool = False,
    ) -> schemas.OperationSet:
        """
        Create and return an Operation set based on the input
        operations and dataset dependencies

        Set param `async_compute` to False to not create op with async

        Args:
            operations: List of operations to add to operation set.
                         Should be in ordered by time operation added.
            dataset_dependency_ids: Dataset ids to set as a parent for this operation set
            async_compute: Set to False not create op set in async fashion in backend/API
            async_verbose: If creating op set in async mode, set verbose to True to have verbose output

        Returns:
            Created Operation Set
        """
        if async_compute:
            # Submit the task request
            task_request = self._operation_set_async(
                operations=operations, dataset_dependency_ids=dataset_dependency_ids
            )
            operation_set_id = polling.poll_operation_set_async_status(task_request=task_request, verbose=async_verbose)
            return self._get._operation_set(operation_set_id)
        else:
            return self._operation_set_non_async(operations=operations, dataset_dependency_ids=dataset_dependency_ids)



"""
API Publish Functions
"""
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal

import numpy as np
import pandas as pd

from pyrasgo import errors
from pyrasgo.primitives.dataset import Dataset, DS_PUBLISHED_STATUS
from pyrasgo.schemas.dataset import DatasetSourceType
from pyrasgo.utils import naming


class Publish:
    """
    API Publish Class
    """

    def __init__(self):
        from pyrasgo.api import Create, Get, Update
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()
        self.create = Create()
        self.update = Update()
        self._dw = None

    @property
    def data_warehouse(self):
        from pyrasgo.storage import DataWarehouse, dw_connection

        if self._dw:
            return self._dw
        self._dw: DataWarehouse = dw_connection(self.api._dw_type())
        return self._dw

    def dataset(
        self,
        dataset: Dataset,
        name: str,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        verbose: bool = False,
        time_index: str = None,
        attributes: Optional[dict] = None,
        table_type: Optional[str] = "VIEW",
        table_name: Optional[str] = None,
        generate_stats: bool = True,
        timeout: Optional[int] = None,
    ) -> Dataset:
        """
        Saves a transformed Dataset in Rasgo to published
        Args:
            dataset: Dataset to save
            name: Name of dataset
            resource_key: A table-safe key used to identify this dataset
            description: Description of dataset
            verbose: If true will print save progress status
            time_index: If the dataset is a time-series with a date column, pass the name of the date column here
            attributes: Dictionary with metadata about the Dataset
            table_type: Type of object to create in snowflake. Can be "TABLE" or "VIEW"
            table_name: Data Warehouse Table Name to set for this DS's published operation
            generate_stats: If True (default) will generate stats for dataset when published
            timeout: Approximate timeout in seconds. Raise an APIError if the dataset isn't available in x seconds
        """
        # Saving of previously-existing Datasets is not allowed
        if dataset._api_dataset:
            raise errors.RasgoRuleViolation(
                f"This Dataset already exists in Rasgo. {dataset}. Transform the dataset to save it."
            )
        if verbose:
            print(f"Saving Dataset with name={name!r} description={description!r} resource_key={resource_key}...")
        operation_set = dataset._get_or_create_op_set()
        if time_index:
            if attributes and "time_index" not in attributes:
                attributes["time_index"] = time_index
            elif attributes and "time_index" in attributes:
                raise errors.ParameterValueError(
                    message=f"Timeseries index explicitly defined as"
                    f" {time_index} in parameters, but defined as "
                    f"{attributes['time_index']} in attributes dict. "
                    f"Please choose to publish either in attributes or "
                    f"as a separate parameter, but not both."
                )
            else:
                attributes = {"time_index": time_index}

        dataset = self.create._dataset(
            name=name,
            source_type=DatasetSourceType.RASGO,
            resource_key=resource_key,
            description=description,
            status=DS_PUBLISHED_STATUS,
            dw_table_id=operation_set.operations[-1].dw_table_id,
            dw_operation_set_id=operation_set.id,
            attributes=attributes,
            publish_ds_table_type=table_type,
            publish_ds_table_name=table_name,
            generate_stats=generate_stats,
            timeout=timeout,
        )
        dataset = Dataset(api_dataset=dataset, api_operation_set=operation_set)
        if verbose:
            print(f"Finished Saving {dataset}")
        return dataset

    def dbt_project(
        self,
        datasets: List[Dataset],
        project_directory: Union[os.PathLike, str] = None,
        models_directory: Union[os.PathLike, str] = None,
        project_name: str = None,
        model_args: Dict[str, Any] = None,
        verbose: bool = False,
    ) -> str:
        """
        Exports all given datasets to Models in a dbt Project

        Params:
        `datasets`: List[Dataset]:
            list of Rasgo datasets to write to dbt as models
        `project_directory`: Path:
            directory to save project files to
            defaults to current working dir
        `models_directory`: Path:
            directory to save model files to
            defaults to project_directory/models
        `project_name`: str:
            name for this project
            defaults to organization name
        """
        from pyrasgo.primitives import DbtProject
        from pyrasgo.utils.dbt import dataset_to_model, dataset_to_source

        project_directory = Path(project_directory or os.getcwd())
        project_name = project_name or naming.cleanse_dbt_name(self.api._profile["organization"]["name"])
        models_directory = Path(models_directory or (project_directory / "models"))
        project = DbtProject(
            name=project_name,
            project_directory=project_directory,
            models_directory=models_directory,
            models=[dataset_to_model(ds) for ds in datasets if not ds.is_source],
            sources=[dataset_to_source(ds) for ds in datasets if ds.is_source],
            model_args=model_args,
        )
        return project.save_files(verbose=verbose)

    def df(
        self,
        df: pd.DataFrame = None,
        name: Optional[str] = None,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        dataset_table_name: str = None,  # TODO: Deprecate in future version
        parents: Optional[List[Dataset]] = None,
        verbose: bool = False,
        attributes: Optional[dict] = None,
        fqtn: Optional[str] = None,
        if_exists: Optional[Literal["append", "overwrite", "fail"]] = "fail",
        generate_stats: bool = True,
    ) -> Dataset:
        """
        Push a Pandas Dataframe a Data Warehouse table and register it as a Rasgo Dataset

        params:
            df: pandas DataFrame
            name: Optional name for the Dataset (if not provided a random string will be used)
            description: Optional description for this Rasgo Dataset
            dataset_table_name: Optional name for the target table (if not provided a random string will be used)
            parents: Set Parent Dataset dependencies for this df dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            fqtn: Optional name for the target table (if not provided a random string will be used)
            if_exists: Values: ['fail', 'append', 'overwrite'] directs the function what to do if a FTQN is passed, and represents an existing Dataset
            generate_stats: If True (default) will generate stats for df dataset when published
        return:
            Rasgo Dataset
        """
        # Make sure no incompatible dw dtypes in df uploading
        _raise_error_if_bad_df_dtypes(df)

        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if_exists_vals = ["overwrite", "append", "fail"]
        if_exists = if_exists.lower()
        if if_exists not in if_exists_vals:
            raise errors.ParameterValueError("if_exists", if_exists_vals)

        if dataset_table_name:
            print(
                "Param `dataset_table_name` will be deprecated in a future version. "
                "Pass `fqtn` to specify the target table for this df."
            )

        target_database, target_schema, target_table = naming.parse_fqtn(
            fqtn or dataset_table_name or naming.random_table_name(), self.api._default_dw_namespace()
        )
        target_fqtn = naming.make_fqtn(database=target_database, schema=target_schema, table=target_table)

        # Check for existing datasets matching FQTN
        ds = None
        try:
            ds = self.get.dataset(fqtn=target_fqtn)
            if ds and verbose:
                print(f"Found Dataset {ds.id} matching FQTN {target_fqtn}. Proceeding in {if_exists} mode.")
        except errors.RasgoResourceException:
            pass

        if ds and if_exists == "overwrite":
            # Users should only be able to overwrite datasets in their own organization
            if ds._api_dataset.organization_id != self.api._profile.get("organizationId"):
                raise errors.RasgoRuleViolation(
                    f"Dataset {ds.id} already exists. This API key does not have permission to replace it."
                )
        elif ds and if_exists == "fail":
            raise errors.ParameterValueError(
                message=f"FQTN {target_fqtn} already exists, and {if_exists} was passed for `if_exists`. "
                "Please confirm the FQTN or choose another value for `if_exists`"
            )

        # Write the df to the target table
        if verbose:
            verb = "Appending" if if_exists == "append" else "Writing"
            print(f"{verb} dataframe to target table {target_fqtn}")
        self.data_warehouse.write_dataframe_to_table(df, table_name=target_fqtn, method=if_exists)

        # Create dataset based on new table FQTN created from df
        if not ds:
            ds = self.table(
                fqtn=target_fqtn,
                name=name or target_fqtn,
                resource_key=resource_key,
                description=description,
                verbose=verbose,
                attributes=attributes,
                parents=parents,
                generate_stats=generate_stats,
                source_type=DatasetSourceType.DATAFRAME,
            )
        return ds

    def table(
        self,
        fqtn: str,
        name: Optional[str] = None,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        parents: Optional[List[Dataset]] = None,
        verbose: bool = False,
        attributes: Optional[dict] = None,
        if_exists: Optional[Literal["return", "update", "fail"]] = "fail",
        generate_stats: bool = True,
        timeout: Optional[int] = None,
        source_type: DatasetSourceType = DatasetSourceType.TABLE,
    ) -> Dataset:
        """
        Register an existing table as a Rasgo Dataset

        params:
            fqtn: The fully qualified table name of the table to register
            name: Optional name to apply to this Rasgo Dataset
            description: Optional description for this Rasgo Dataset
            parents: Set Parent Dataset dependencies for this table dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            generate_stats: If True (default) will generate stats for table dataset when published
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
            source_type: Specifies the `source_type` of this Dataset in Rasgo. Defaults to 'TABLE'
        return:
            Rasgo Dataset
        """
        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if verbose:
            print(f"Publishing {source_type.value} {fqtn} as Rasgo dataset")

        if fqtn.count(".") != 2:
            raise errors.ParameterValueError(
                message=f"'{fqtn}' is not a valid fully qualified table name. "
                "FQTNs should follow the format DATABASE.SCHEMA.TABLE_NAME.  "
                "Please pass a valid FQTN and try again"
            )

        table_database, table_schema, table_name = naming.parse_fqtn(fqtn)

        try:
            row_count = self.data_warehouse.query_into_dict(
                f"select count(1) as ROW_CT from {table_database}.{table_schema}.{table_name}"
            )
            if row_count[0]["ROW_CT"] == 0:
                raise errors.DWResourceException(
                    f"Source table {table_name} is empty or this role does not have access to it."
                )
        except Exception as err:
            raise errors.DWResourceException(
                f"Source table {table_name} does not exist or this role does not have access to it."
            ) from err

        # Make sure `source_type` param is valid Enum Value
        if not isinstance(source_type, DatasetSourceType):
            raise errors.ParameterValueError("source_type", ["CSV", "RASGO", "TABLE", "DATAFRAME"])

        # Check if a Dataset already exists
        existing_ds = None
        try:
            existing_ds = self.get.dataset(fqtn=fqtn)
        except errors.RasgoResourceException:
            pass
        if existing_ds:
            if if_exists == "return":
                return existing_ds
            elif if_exists == "update":
                dataset = self.update.dataset(
                    dataset=existing_ds,
                    name=name or table_name,
                    description=description,
                    attributes=attributes,
                )
                return Dataset(api_dataset=dataset)
            else:  # if_exists == "fail":
                raise errors.RasgoResourceException(
                    f"{fqtn} is already registered with Rasgo as Dataset {existing_ds.id}"
                )

        # Create operation set with parent dependencies
        # set for this dataset
        operation_set = self.create._operation_set(
            operations=[],
            dataset_dependency_ids=parent_ids,
            async_compute=False,
        )

        # Publish Dataset with operation set created above
        dataset = self.create._dataset(
            name=name or table_name,
            source_type=source_type,
            resource_key=resource_key,
            description=description,
            fqtn=fqtn,
            status=DS_PUBLISHED_STATUS,
            attributes=attributes,
            dw_operation_set_id=operation_set.id,
            generate_stats=generate_stats,
            timeout=timeout,
        )
        # Raise API error if backend error creating dataset
        if not dataset:
            raise errors.APIError("DataSource failed to upload")

        # Return dataset if no error
        return Dataset(api_dataset=dataset)


# ------------------------------------------
#  Private Helper Funcs for Publish Class
# ------------------------------------------


def _raise_error_if_bad_df_dtypes(df: pd.DataFrame) -> None:
    """
    Raise an API error is any dtypes in the pandas dataframe,
    which are being pushed to the data warehouse aren't compatible.

    Raise proper error message if so
    """
    for col_name in df:
        col = df[col_name]
        if col.dtype.type == np.datetime64:
            raise errors.RasgoRuleViolation(
                "Can't publish pandas Df to Rasgo. Df column "
                f"'{col_name}' needs to be converted to proper datetime format.\n\n"
                "If your column is a **DATE** use `pd.to_datetime(df[<col_name>]).dt.date` to convert it\n"
                "If your column is a **TIMESTAMP** use `pd.to_datetime(final_df['col_name']).dt.tz_localize('UTC')`"
            )



from enum import Enum
from typing import List, Optional
from typing_extensions import Literal

import pandas as pd

from pyrasgo import errors
from pyrasgo.utils import naming


def build_schema(df: pd.DataFrame, include_index=False) -> dict:
    """
    Returns a dict representing the schema of a dataframe
    """
    from pandas.io.json import build_table_schema

    schema_list = build_table_schema(df)
    if not include_index:
        return {column['name']: column for column in schema_list['fields'] if column['name'] != 'index'}
    return {column['name']: column for column in schema_list['fields']}


def generate_ddl(
    df: pd.DataFrame,
    table_name: str,
    method: Literal["append", "overwrite", "fail"] = None,
):
    """
    Generates a SQL statement to create a table matching the schema of a dataframe
    """
    create_statement = "CREATE TABLE IF NOT EXISTS"
    if method == "overwrite":
        create_statement = "CREATE OR REPLACE TABLE"
    return pd.io.sql.get_schema(df, table_name).replace("CREATE TABLE", create_statement).replace('"', '')


def confirm_df_columns(
    df: pd.DataFrame,
    dimensions: List[str],
    features: List[str] = None,
):
    """
    Checks whether a list of dimensions and features are in a dataframe
    then returns valid columns
    """
    return confirm_list_columns(list(df.columns), dimensions, features or [])


def confirm_list_columns(
    columns: list,
    dimensions: List[str],
    features: List[str] = None,
):
    """
    Checks whether a list of dimensions and features are in a list of columns
    then returns valid columns
    """
    missing_dims = []
    missing_features = []
    valid_dims = []
    valid_features = []
    consider = []
    for dim in dimensions:
        if dim not in columns:
            missing_dims.append(dim)
            if naming.cleanse_sql_name(dim) in columns:
                consider.append(naming.cleanse_sql_name(dim))
        else:
            valid_dims.append(dim)
    if not features:
        features = set(columns) - set(valid_dims)
    for ft in list(features):
        if ft not in columns:
            missing_features.append(ft)
            if naming.cleanse_sql_name(ft) in columns:
                consider.append(naming.cleanse_sql_name(ft))
        else:
            valid_features.append(ft)
    if missing_dims or missing_features:
        raise errors.DWResourceException(
            "Specified columns do not exist in dataframe: "
            f"Dimensions({missing_dims}) Features({missing_features}) "
            f"Consider these: ({consider})?"
        )
    return valid_dims, valid_features


def generate_unique_id() -> str:
    """
    Calculates a unique 32 char string that can be used in secure urls
    """
    import secrets

    return secrets.token_urlsafe(32)


def map_pandas_df_type(pandas_df_type: str) -> str:
    """For a given pandas dataframe type, return the equivalent python type.
    Default to input if no match"""
    if pandas_df_type.startswith('float'):
        return 'float'
    if pandas_df_type.startswith('datetime'):
        return 'datetime'
    if pandas_df_type.startswith('int'):
        return 'integer'
    return pandas_df_type


def set_df_id(df: pd.DataFrame, experiment_id: Optional[str]) -> int:
    """
    Checks whether a df is part of an ongoing experiment and tags it with
    the experiment id or a new id
    """
    # If ID is set via experiment, use it
    if experiment_id:
        tag_dataframe(df, {"RasgoID": experiment_id})
        return experiment_id
    # If df is already marked with an ID, keep it
    elif df.attrs.get("RasgoID"):
        return df.attrs["RasgoID"]
    # Else calculate a new ID
    else:
        unique_id = generate_unique_id()
        tag_dataframe(df, {"RasgoID": unique_id})
        return unique_id


def tag_dataframe(df: pd.DataFrame, attribute: dict):
    """
    Tags a dataframe with a user-defined attribute
    """
    write_back_attrs = df.attrs or {}
    write_back_attrs.update(attribute)
    df.attrs = write_back_attrs


def cleanse_sql_dataframe(df: pd.DataFrame):
    """
    Renames all columns in a pandas dataframe to Snowflake compliant names in place
    """
    df.rename(columns={r: naming.cleanse_sql_name(r) for r in build_schema(df)}, inplace=True)



"""
Manage Data Warehouse specific imports
"""

### ------------
# Note to users:
# This module is used to allow pyrasgo to install the minimum dependencies
# needed to run your flavor of DataWarehouse

# We support downloading extas packages for each DW: e.g. `pyrasgo[snowflake]`
# so your machine is not cluttered with unused google packages.
# In doing so, we run the risk of not having required dependencies at runtime

# This module attempts to import all possible datawarehouse packages, and
# fails gracefully if they are not present
# pyRasgo will use these empty aliases to raise a PackageDependencyWarning
# from the class or function that needs the import
### ------------

# Attempt safe imports of all DW packages so we can warn later if they are missing
try:
    from snowflake import connector as sf_connector
    from snowflake.connector.pandas_tools import write_pandas
except ImportError:
    sf_connector = None
    write_pandas = None

try:
    from google.cloud import bigquery as bq
    from google.cloud import exceptions as gcp_exc
    from google.oauth2 import service_account as gcp_svc
except ImportError:
    bq = None
    gcp_exc = None
    gcp_svc = None

try:
    from google_auth_oauthlib import flow as gcp_flow
except ImportError:
    gcp_flow = None



from typing import List, Optional

from pydantic import BaseModel


class SimpleColumn(BaseModel):
    columnName: str
    dataType: str


class SimpleTable(BaseModel):
    tableName: str
    databaseName: str
    schemaName: str
    fqtn: Optional[str]


class SimpleTableWithColumns(SimpleTable):
    columns: Optional[List[SimpleColumn]]



"""
API Read Functions
"""
from typing import List, Optional

import pandas as pd

from pyrasgo import errors
from pyrasgo.primitives import Dataset


class Read:
    """
    API Read Class
    """

    def __init__(self):
        from pyrasgo.api import Get
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()
        self._dw = None

    @property
    def data_warehouse(self):
        from pyrasgo.storage import DataWarehouse, dw_connection

        if not self._dw:
            self._dw: DataWarehouse = dw_connection(self.api._dw_type())
        return self._dw

    def dataset(
        self,
        id: Optional[int] = None,
        dataset: Optional[Dataset] = None,
        filters: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
        limit: Optional[int] = None,
        snapshot_index: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        Constructs and returns pandas DataFrame from the specified Rasgo Dataset

        You can supply SQL WHERE clause filters, order the dataset by columns, only
        return selected columns, and add a return limit as well

        Example:
            ```
            rasgo.read.dataset(
                id=74,
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY'],
                limit=50
            )
            ```

        Args:
            id: dataset id to read into df
            dataset: Dataset obj to read into df
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
            limit: Only return this many rows in the df
            snapshot: the index of a snapshot from Dataset.snapshots to read
        """
        # Validate one dataset passed in, id or dataset obj
        if not dataset and id is None:
            raise errors.ParameterValueError(
                message="Must pass either a valid dataset ID or Dataset object to read into a DataFrame"
            )
        if not dataset:
            # Note: Func below already raises API error if dataset with id doesn't exist
            dataset = self.get.dataset(id)

        # Require the operation set on the DS to make
        # sure table is created before reading
        dataset._get_or_create_op_set()

        # Get the FQTN from snapshot or current dataset
        if snapshot_index:
            try:
                fqtn = dataset._api_dataset.snapshots[snapshot_index].fqtn
            except IndexError as err:
                raise errors.RasgoResourceException(f"Snapshot index {snapshot_index} does not exist") from err
        else:
            fqtn = dataset.fqtn

        if dataset:
            try:
                query = self.data_warehouse.make_select_statement(
                    table_metadata={'fqtn': fqtn},
                    filters=filters,
                    order_by=order_by,
                    columns=columns,
                    limit=limit,
                )
                return self.data_warehouse.query_into_dataframe(query)
            except Exception as err:
                raise errors.DWResourceException(f"Dataset table is not reachable: {err}") from err



pyrasgo_api_key: str = None

# Set max polling time to an hour
MAX_POLL_ATTEMPTS = 900
POLL_RETRY_RATE = 4


def get_session_api_key():
    global pyrasgo_api_key
    return pyrasgo_api_key


def set_session_api_key(api_key: str):
    global pyrasgo_api_key
    clear_session_api_key()
    pyrasgo_api_key = api_key


def clear_session_api_key():
    global pyrasgo_api_key
    pyrasgo_api_key = None



import json
from typing import Dict
from urllib.parse import quote, urlencode

import requests
from requests import Response

from pyrasgo import errors
from pyrasgo.api.session import SessionMeta, generate_headers


class Connection(object, metaclass=SessionMeta):
    """
    Base class for all Rasgo objects to facilitate API calls.
    """

    def __init__(self, api_key=None):
        if api_key:
            self._api_key = api_key

    def _default_dw_namespace(self) -> Dict:
        dw_creds = self._profile.get("dw_creds")
        organization = self._profile.get("organization")
        if not dw_creds or not organization:
            raise errors.DWCredentialsWarning("Your user profile is missing critical information. Contact Rasgo.")
        return {
            'database': organization.get("database", dw_creds.get("project")),
            'schema': organization.get("schema", dw_creds.get("dataset")),
        }

    def _dw_type(self) -> str:
        org = self._profile.get('organization')
        if not org:
            raise errors.DWCredentialsWarning("Your user profile is missing critical information. Contact Rasgo.")
        return org.get('cloudDataWarehouse', org.get('dw_type'))

    def _url(self, resource, api_version=None):
        if '/' == resource[0]:
            resource = resource[1:]
        protocol = 'http' if self._environment.value == 'localhost' else 'https'
        return f"{protocol}://{self._environment.value}/{'' if api_version is None else f'v{api_version}/'}{resource}"

    def find(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filter_str = k + "||$eq||" + v
            search_strings.append(filter_str)
        params = {"filter": search_strings}
        params = urlencode(params, quote_via=quote, doseq=True)
        response = self._get(resource, params)
        if response.status_code == 404:
            return None
        else:
            self._raise_internal_api_error_if_any(response)
            return response.json()

    def _delete(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs DELETE request to Rasgo API as defined within the class instance.

        :param resource: Target resource to DELETE from API.
        :param _json: JSON object to send in DELETE request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.delete(
            self._url(resource, api_version),
            json=_json,
            headers=generate_headers(self._api_key),
            params=params or {},
        )
        self._raise_internal_api_error_if_any(response)
        return response

    def _get(self, resource, params=None, api_version=None) -> requests.Response:
        """
        Performs GET request to Rasgo API as defined within the class instance.

        :param endpoint: Target resource to GET from API.
        :param params: Additional parameters to specify for GET request.
        :return: Response object containing content returned.
        """
        response = requests.get(
            self._url(resource, api_version),
            headers=generate_headers(self._api_key),
            params=params or {},
        )
        self._raise_internal_api_error_if_any(response)
        return response

    def _patch(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PATCH request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PATCH from API.
        :param _json: JSON object to send in PATCH request
        :param params: Additional parameters to specify for PATCH request.
        :return: Response object containing content returned.
        """
        response = requests.patch(
            self._url(resource, api_version),
            json=_json,
            headers=generate_headers(self._api_key),
            params=params or {},
        )
        self._raise_internal_api_error_if_any(response)
        return response

    def _post(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs POST request to Rasgo API as defined within the class instance.

        :param resource: Target resource to POST from API.
        :param _json: JSON object to send in POST request
        :param params: Additional parameters to specify for POST request.
        :return: Response object containing content returned.
        """
        response = requests.post(
            self._url(resource, api_version),
            json=_json,
            headers=generate_headers(self._api_key),
            params=params or {},
        )
        self._raise_internal_api_error_if_any(response)
        return response

    def _put(self, resource, _json=None, params=None, api_version=None) -> requests.Response:
        """
        Performs PUT request to Rasgo API as defined within the class instance.

        :param resource: Target resource to PUT from API.
        :param _json: JSON object to send in PUT request
        :param params: Additional parameters to specify for PUT request.
        :return: Response object containing content returned.
        """
        response = requests.put(
            self._url(resource, api_version),
            json=_json,
            headers=generate_headers(self._api_key),
            params=params or {},
        )
        self._raise_internal_api_error_if_any(response)
        return response

    def _get_items(self, resource, equality_filters):
        search_strings = []
        for k, v in equality_filters.items():
            filter_str = k + "||$eq||" + v
            search_strings.append(filter_str)
        params = {"filter": search_strings}
        params = urlencode(params, quote_via=quote, doseq=True)
        try:
            response = self._get(resource, params)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise e
        if response.status_code == 404:
            return None
        else:
            self._raise_internal_api_error_if_any(response)
            return response.json()

    def _get_item(self, resource, equality_filters, allow_multiple=False):
        results = self._get_items(resource, equality_filters)
        if results is None:
            return None
        elif 1 < len(results):
            if allow_multiple:
                return results[-1]
            else:
                raise errors.APIError(f"Multiple returned for {resource}, check your filters")
        elif 0 == len(results):
            return None
        else:
            return results[0]

    def _raise_internal_api_error_if_any(self, response: Response) -> None:
        """
        Raise an internal API error if any occurred

        Show proper APIError Message in PyRasgo by parsing
        FastAPI HttpError msg
        """
        try:
            response.raise_for_status()
        except Exception as err:
            try:
                err_details = response.json()
            except json.JSONDecodeError:
                err_details = response.text
            raise errors.APIError(
                f"Internal API Error when making {response.request.method} request\n\n"
                f"Status Code: {response.status_code}\n\n"
                f"Internal Error Details: {err_details}"
            ) from err



"""
Dataset 'Primitive' In Rasgo SDK
"""
from __future__ import annotations
import functools
import inspect
from datetime import datetime
from inspect import Parameter
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from pyrasgo import errors
from pyrasgo.api.connection import Connection
from pyrasgo.schemas import (
    OperationSet,
    Dataset as DatasetSchema,
    OperationCreate,
    Transform,
    DatasetBulk,
    DatasetColumn,
    DatasetSourceType,
)
from pyrasgo.storage.datawarehouse import dw_connection
from pyrasgo.utils import naming
from pyrasgo.utils import exports
from pyrasgo.constants import DS_PUBLISHED_STATUS, SOURCE_TABLE_ARG_NAME, TABLE_ARG_TYPE, TABLE_LIST_ARG_TYPE

__all__ = ['Dataset']


def require_published(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: Dataset = args[0]
        self._assert_is_published()
        return func(*args, **kwargs)

    return wrapper


def require_operations(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: Dataset = args[0]
        self._assert_has_operations()
        return func(*args, **kwargs)

    return wrapper


class Dataset(Connection):
    """
    Representation of a Rasgo Dataset
    """

    def __init__(
        self,
        # Args passed from Rasgo
        api_dataset: Optional[DatasetSchema] = None,
        api_operation_set: Optional[OperationSet] = None,
        # Args passed from transforming
        operations: Optional[List[OperationCreate]] = None,
        dataset_dependencies: Optional[List[int]] = None,
        table_name: Optional[str] = None,
        transforms: Optional[List[Transform]] = None,
        verbose=False,
        **kwargs: Dict,
    ):
        """
        Init functions in two modes:
            1. This Dataset retrieved from Rasgo. This object is for reference, and cannot
               be changed, but can be transformed to build new datasets
            2. This Dataset represents a new dataset under construction. It is not persisted in Rasgo
               and instead consists of some operations that will be used to generate a new dataset.
        """
        super().__init__(**kwargs)

        self._verbose = verbose
        self._api_dataset: DatasetSchema = api_dataset
        self._api_operation_set: OperationSet = api_operation_set
        self._operations = operations if operations else []
        self._dataset_dependencies = dataset_dependencies if dataset_dependencies else []
        self._table_name = table_name
        if transforms:
            self._available_transforms = transforms
        else:
            self._available_transforms = _get_transforms()

        #  alias .transform allowing direct referencing of named transforms
        for transform in self._available_transforms:
            f = self._create_transform_function(transform)
            setattr(self, transform.name, f)

    def __repr__(self) -> str:
        """
        Get string representation of this dataset
        """
        if self._api_dataset:
            return (
                f"Dataset(id={self.id}, "
                f"name={self.name}, "
                f"resource_key={self._api_dataset.resource_key}, "
                f"version={self._api_dataset.version}, "
                f"status={self.status}, "
                f"description={self.description})"
            )
        else:
            return "Dataset()"

    # -------------------
    # Properties
    # -------------------

    @property
    def id(self) -> Optional[int]:
        """
        Return the id for this dataset

        Raise API error if one doesn't exist yet and is an offline dataset
        """
        if self._api_dataset:
            return self._api_dataset.id

    @property
    def name(self) -> Optional[str]:
        """
        Return dataset name if set/saved
        """
        if self._api_dataset:
            return self._api_dataset.name

    @property
    def description(self) -> Optional[str]:
        """
        Return dataset description if set/saved
        """
        if self._api_dataset:
            return self._api_dataset.description

    @property
    def status(self) -> Optional[str]:
        """
        Return dataset status
        """
        if self._api_dataset:
            return self._api_dataset.status

    @property
    def fqtn(self) -> str:
        """
        Returns the Fully Qualified Table Name for this dataset
        """
        if self._api_dataset and self._api_dataset.dw_table:
            return self._api_dataset.dw_table.fqtn
        elif self._table_name:
            dw_namespace = self._default_dw_namespace()
            return f"{dw_namespace['database']}.{dw_namespace['schema']}.{self._table_name}"
        else:
            raise AttributeError("No fqtn exists for this Dataset")

    @property
    def table_type(self) -> Optional[str]:
        """
        If this is a published dataset, return the
        table type for it's set Dw Table ( 'VIEW' or 'TABLE')
        """
        if self._api_dataset and self._api_dataset.dw_table:
            return self._api_dataset.dw_table.table_type
        return "VIEW"

    @property
    def columns(self) -> Optional[List[DatasetColumn]]:
        """
        Return the columns for this dataset if it is from the API
        """
        if self._api_dataset:
            # If the dataset was from bulk `rasgo.get.datasets()`
            # it won't have dataset columns, so make API call
            # to retrieve them.
            if self._api_dataset.columns is None:
                from pyrasgo.api import Get

                self._api_dataset.columns = Get()._dataset_columns(self.id)

            return self._api_dataset.columns

    @property
    def resource_key(self) -> Optional[str]:
        """
        Return the resource key for this dataset.

        Only published datasets will contain this value
        """
        if self._api_dataset:
            return self._api_dataset.resource_key

    @property
    def version(self) -> Optional[int]:
        """
        Return the version for this dataset.

        Only published datasets will contain this value
        """
        if self._api_dataset:
            return self._api_dataset.version

    @property
    def created_date(self) -> Optional[datetime]:
        """
        Return date this dataset was created
        """
        if self._api_dataset:
            return self._api_dataset.create_timestamp

    @property
    def update_date(self) -> Optional[datetime]:
        """
        Return date this dataset was updated
        """
        if self._api_dataset:
            return self._api_dataset.update_timestamp

    @property
    def attributes(self) -> Optional[Dict]:
        """
        Return the attributes for this dataset if it is from the API
        """
        if self._api_dataset:
            return self._api_dataset.attributes

    @property
    def source_type(self) -> Optional[str]:
        """
        Return dataset source type if set on published Dataset

        All offline datasets are of type `RASGO`
        """
        if self._api_dataset:
            return self._api_dataset.source_type.value
        else:
            return DatasetSourceType.RASGO.value

    @property
    def is_source(self) -> bool:
        """
        Is this Dataset a standalone source? i.e. does it have no applied operations?
        Sources are imported from csvs, dataframes or directly from Snowflake tables

        This property is a convenience calc to support exporting Datasets to dbt
        """
        if self._api_dataset:
            return self._api_dataset.is_source

    @property
    def dependencies(self) -> List[Dataset]:
        """
        Return a list of dataset dependencies for this dataset
        """
        from pyrasgo.api import Get

        dataset_deps = []
        self._refresh()
        if self._api_operation_set:
            for ds in self._api_operation_set.dataset_dependencies:
                dataset_deps.append(Get().dataset(ds.id))
        else:
            for ds_id in self._dataset_dependencies:
                dataset_deps.append(Get().dataset(ds_id))
        return dataset_deps

    @property
    def sql(self) -> Optional[str]:
        """
        Return the source code SQL used to generate this dataset
        """
        return self._get_cte_for_dataset()

    @property
    def snapshots(self) -> Optional[list]:
        """return a list of tuples of timestamp identified by index and the snapshot creation timestamp"""
        return [(i + 1, ss.timestamp) for i, ss in enumerate(self._api_dataset.snapshots)]

    @property
    def versions(self) -> List[DatasetBulk]:
        """return a list of versions of this dataset"""
        return self._get(
            f"datasets/rk/{self._api_dataset.resource_key}/versions",
            api_version=2,
        ).json()

    @property
    @require_published
    def schema(self) -> Optional[List[Tuple[str, str]]]:
        """
        Return Dataset Schema has list of LIst of Tuples like (<Col Name>, <Col Data Type>)
        """
        if self._api_dataset:
            data_warehouse = dw_connection(self._dw_type())
            return data_warehouse.list_table_columns(self._api_dataset.dw_table.fqtn)

    # --------
    # Methods
    # --------

    def transform(
        self,
        transform_name: str,
        arguments: Optional[Dict[str, Union[str, int, List, Dict, Dataset]]] = None,
        operation_name: Optional[str] = None,
        render_only: Optional[bool] = False,
        **kwargs: Union[str, int, List, Dict, Dataset],
    ) -> Union[Dataset, None]:
        """
        Transform a new dataset with the given transform and arguments.
        Created operation is added to the dataset's canvas/operations set

        Args:
            transform_name: Name of transform to Apply
            arguments: Optional transform arguments sin not supplied by **kwargs
            operation_name: Name to set for the operation/transform
            render_only: Optional flag to simply render the operation that will result
                from this transformation instead of using it to create a new Dataset.
            **kwargs:

        Returns:
             Returns an new dataset with the referenced transform
             added to this dataset's definition/operation set

             Optionally, can be used to try out some inputs to a transform by passing `render_only = True`
        """
        # Update the Transform arguments with any supplied kwargs
        arguments = arguments if arguments else {}
        arguments.update(kwargs)

        # Do any validation on input transform args
        transform = self._get_transform_by_name(transform_name)
        _assert_value_types_of_args(transform, arguments, self)

        # Add required reference to self in transform
        arguments[SOURCE_TABLE_ARG_NAME] = self

        # Scan through supplied transform arguments
        #
        # Determine what should be set for this operation's plus offline dataset's
        # dependencies, along with getting the parent operations for this operation
        # by reading data from input Datasets obj attribute data
        #
        # We also convert 'Datasets' to fqtns in the arguments dict in
        # the func below
        op_deps, ds_deps, parent_ops, arguments = self._get_op_deps_parents_and_args(arguments)

        # Init table name for outputted dataset
        table_name = naming.gen_operation_table_name(
            op_num=len(parent_ops) + 1,
            transform_name=transform_name,
        )

        # Init New Operation Create Contract
        operation_create = OperationCreate(
            operation_name=operation_name if operation_name else transform.name,
            operation_args=arguments,
            transform_id=transform.id,
            table_name=table_name,
            table_dependency_names=op_deps,
        )

        # If rendering only, short circuit and get the SQL we'll
        # render for this operation
        if render_only:
            from pyrasgo.api import Create

            print(Create()._operation_render(operation_create))
            return

        # Init and return new offline dataset
        return self.__class__(
            operations=parent_ops + [operation_create],
            dataset_dependencies=ds_deps,
            table_name=table_name,
        )

    def to_df(
        self,
        filters: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
        limit: Optional[int] = None,
        snapshot_index: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        Reads and returns this dataset into a pandas dataframe

        You can supply SQL WHERE clause filters, order the dataset by columns, only
        return selected columns, and add a return limit as well

        Example:
            ```
            ds = rasgo.get.dataset(dataset_id=74)
            ds.to_df(
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY'],
                limit=50
            )
            ```

        Args:
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
            limit: Only return this many rows in the df
            snapshot_index: the index of a snapshot from Dataset.snapshots to read
        """
        from pyrasgo.api import Read
        from pyrasgo.utils.rendering import offline_operations_as_cte

        if self._operations:
            sql = offline_operations_as_cte(operations=self._operations, transforms=self._available_transforms)
            if filters:
                sql = f"{sql} WHERE {' AND '.join(filters)}"
            if order_by:
                sql = f"{sql} ORDER BY {', '.join(order_by)}"
            if isinstance(limit, int):
                sql = f"{sql} LIMIT {limit}"
            if columns:
                final_op_name = naming.gen_operation_table_name(
                    op_num=len(self._operations) + 1,
                    transform_name='drop_columns',
                )
                sql = f"WITH {final_op_name} AS ({sql}) SELECT {','.join(columns)} FROM {final_op_name}"
            return Read().data_warehouse.query_into_dataframe(sql)
        return Read().dataset(
            dataset=self,
            filters=filters,
            order_by=order_by,
            columns=columns,
            limit=limit,
            snapshot_index=snapshot_index - 1 if snapshot_index else None,
        )

    def preview(
        self,
        filters: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """
        Preview the first 10 rows of this dataset, returned as pandas dataframe

        You can supply SQL WHERE clause filters, order the dataset by columns, and
        only return selected columns

        Example:
            ```
            ds = rasgo.get.dataset(dataset_id=74)
            ds.preview(
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY']
            )
            ```

        Args:
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
        """
        return self.to_df(filters, order_by, columns, limit=10)

    def generate_yaml(self, file_path: Optional[Path] = None) -> str:
        """
        Return a YAML representation of this dataset

        Args:
            file_path: str: full path + file name of the yaml file to write to
        """
        from pyrasgo.api import Get

        ds_schema = Get().dataset_offline_version(self.resource_key, self.version)
        if file_path:
            with open(file_path, "w") as yaml_file:
                yaml_file.write(ds_schema.yaml())
            print(f"Dataset version written to yaml file at: {file_path}")
        return ds_schema.yaml()

    @require_published
    def generate_py(self) -> str:
        """
        Generate and return as a string the PyRasgo code which
        will create an offline a copy of this dataset whether DS
        is in draft or unpublished status.

        Dataset must be published in Rasgo first before you can call this func.
        """
        from pyrasgo.api import Get

        return Get().dataset_py(self.id)

    @require_published
    def dw_sync(self) -> None:
        """
        Looks up the schema of this Dataset's table in the DataWarehouse and synchronizes
        the Dataset's column metadata to match it

        Helpful to pick up common changes:
        - new fields added to table
        - existing fields removed from table
        - data type changes

        Does not pick up:
        - table renames
        - column renames

        Dataset must be published in Rasgo first before you can call this func.
        """
        from pyrasgo.api import Update

        Update().dataset_schema(self.id)
        self._refresh()

    def profile(self) -> any:
        """
        Get the URL for this Dataset in the RasgoUI
        """
        self._assert_is_created()
        from pyrasgo.api.session import Environment

        return f"{Environment.from_environment().app_path}/datasets/{self.id}/canvas"

    @require_published
    def run_stats(self, only_if_data_changed: Optional[bool] = True) -> None:
        """
        Trigger new stats for this dataset

        Args:
            only_if_data_changed: Use False to tell Rasgo to run stats even if the data has not changed

        Dataset must be published in Rasgo first before you can call this func.
        """
        from pyrasgo.api.create import Create

        Create()._dataset_correlation_stats(
            table_id=self._api_dataset.dw_table_id, only_if_data_changed=only_if_data_changed
        )
        print(
            f"Request to generate stats for Dataset {self.id} started. Once generated, view stats at {self.profile()}"
        )

    @require_published
    def refresh_table(self, verbose: bool = False, timeout: Optional[int] = None) -> None:
        """
        Kicks off a query for re-materializing this dataset's set Dw Table.

        The dataset needs to be published first in order call this function

        Args:
            verbose: If True will print information related to refreshing table
            timeout: Approximate timeout for creating the table in seconds.
                     Raise an APIError if timeout reached

        Dataset must be published in Rasgo first before you can call this func.
        """
        from pyrasgo.api.update import Update

        Update().dataset_table(dataset=self, verbose=verbose, timeout=timeout)

    @require_published
    def take_snapshot(self) -> List:
        """
        Take a snapshot of this dataset
        """
        from pyrasgo.api import Create

        Create().dataset_snapshot(self.id)
        self._refresh()
        return self.snapshots

    def to_sql(self) -> Optional[str]:
        """
        Return the SQL statement to create this dataset
        """
        sql = self._get_cte_for_dataset()
        if sql:
            return f"CREATE OR REPLACE {self.table_type} {self.fqtn} AS {sql}"
        return self.sql

    def to_accelerator(self) -> Optional[str]:
        if self._api_dataset:
            return exports.dataset_to_accelerator_yaml(dataset=self)

    # ---------------------------------
    #  Private Helper Funcs for Class
    # ---------------------------------

    def _refresh(self) -> None:
        """
        Refresh this Dataset's metadata from the Rasgo API
        """
        if self.id:
            self._api_dataset = DatasetSchema(**self._get(f"/datasets/{self.id}", api_version=2).json())
        if self._api_dataset:
            self._api_operation_set = OperationSet(
                **self._get(f"/operation-sets/{self._api_dataset.dw_operation_set_id}", api_version=2).json()
            )

    def _get_cte_for_dataset(self) -> str:
        """
        Return the SQL statement to create this dataset as a CTE
        """
        from pyrasgo.utils.rendering import operations_as_cte, offline_operations_as_cte

        # Need to pull the operation set in case this Dataset was sourced from a bulk contract
        if not self._api_operation_set:
            self._refresh()
        # If this is an Online Dataset with operations:
        if self._api_operation_set and self._api_operation_set.operations:
            return operations_as_cte(self._api_operation_set.operations)
        # If this is a published Dataset without operations:
        if self._api_dataset and self._api_dataset.dw_table:
            return f'SELECT * FROM {self._api_dataset.dw_table.fqtn}'
        # If this is an Offline Dataset:
        # For now: return None | Future: build an offline CTE rendering function
        else:
            return offline_operations_as_cte(operations=self._operations, transforms=self._available_transforms)

    def _get_transform_by_name(self, transform_name: str) -> Transform:
        """
        Get and return a transform obj by name

        Raise Error if no transform with that name found
        """
        for transform in self._available_transforms:
            if transform_name == transform.name:
                return transform
        raise errors.RasgoResourceException(f"No Transform with name '{transform_name}' available to your organization")

    def _assert_can_transform(self) -> None:
        """
        Raise an API error if you can't transform this dataset
        """
        if self._api_dataset:
            if not self._api_dataset.dw_table_id or self.status != DS_PUBLISHED_STATUS:
                raise errors.RasgoRuleViolation(
                    f"Dataset({self._api_dataset.id}) has not been published, "
                    f"and cannot be used as an arg for new dataset transformations"
                )

    def _assert_is_created(self) -> None:
        """
        Raise an error if the dataset is not created in Rasgo
        """
        if not self._api_dataset:
            raise errors.RasgoRuleViolation("Dataset must be registered with Rasgo order to perform this action.")

    def _assert_is_published(self) -> None:
        """
        Raise an error if the dataset is not published
        """
        if self.status != DS_PUBLISHED_STATUS or not self._api_dataset.dw_table_id:
            raise errors.RasgoRuleViolation(
                "Dataset must be published in order to perform this action. Please publish your dataset to continue."
            )

    def _assert_has_operations(self) -> None:
        """
        Raise an error if the dataset does not have transforms applied
        """
        ops = None
        ops = self._api_operation_set.operations if self._api_operation_set else self._operations
        if not ops:
            raise errors.RasgoRuleViolation(
                "Dataset must have transforms applied in order to perform this action. "
                "Please apply transforms to your dataset to continue. "
            )

    def _get_or_create_op_set(self) -> OperationSet:
        """
        Get or create plus return the operation set for this dataset
        """
        # Online Dataset: Try to get existing OperationSet from the API first
        if not self._api_operation_set:
            self._refresh()

        # Offline Dataset: If there is no OpSet available in the API, create one
        if not self._api_operation_set:
            from pyrasgo.api.create import Create

            self._api_operation_set = Create()._operation_set(
                operations=self._operations,
                dataset_dependency_ids=self._dataset_dependencies,
                async_verbose=self._verbose,
            )

        return self._api_operation_set

    def _create_transform_function(self, transform: Transform) -> Callable:
        """
        Creates and returns a new function to dynamically attached to the Dataset obj on init

        New funcs docstring, name, and signature (params shown when inspecting/doing . tab on func)
        as well to improve notebook experience of using transforms for users

        Args:
            ds_transform_func: Function pointer of Dataset.transform()
            transform: Transform to read metadata and create new function for
        """

        # Create new function with 'transform_name` param set to this transform's name
        def f(*args, **kwargs) -> Dataset:
            return self.transform(transform_name=transform.name, *args, **kwargs)

        # Update func meta data for better inspection in notebook
        f.__name__ = transform.name
        f.__signature__ = _gen_func_signature(f, transform)
        f.__doc__ = _gen_func_docstring(transform)
        return f

    def _get_op_deps_parents_and_args(
        self,
        arguments: Dict[str, Union[str, int, List, Dict, Dataset]],
    ) -> Tuple[List[str], List[int], List[OperationCreate], Dict[str, Union[str, int, List, Dict, Dataset]]]:
        """
        Based on the input arguments a user supplied for this transform,
        determine and return a many things needed to initialize the next
        offline dataset. This includes
          - Operation Dependencies as list of FQTNs
          - Dataset Dependencies as list of unique dataset ids
          - List of parent operations for the next offline dataset
          - Updated args dict with Dataset objs converted to FQTN stings

        Args:
            arguments: The arguments a user supplied to the transform. This includes
                       'source_table' value we auto-populate

        Returns:
            Tuple of things needed to create next offline dataset
        """
        # Make copy of supplied transform arguments
        # since modifying in function
        arguments = arguments.copy()

        # Init lists to keep track of this operation's deps,
        # parent operations along with op set dependencies
        op_deps = []
        parent_ops = []
        ds_deps = self._dataset_dependencies.copy()

        # Always handle the `source_table` first to keep proper
        # order of parent operations
        source_table: Dataset = arguments.pop(SOURCE_TABLE_ARG_NAME)
        op_deps, parent_ops, ds_deps = self._update_op_deps__ds_deps__and_parent_ops(
            ds_arg=source_table, op_deps=op_deps, parent_ops=parent_ops, ds_deps=ds_deps
        )
        arguments[SOURCE_TABLE_ARG_NAME] = source_table.fqtn

        arguments, new_op_deps, new_parent_ops, new_ds_deps = self._replace_ds_args(arguments)
        op_deps += new_op_deps
        parent_ops += new_parent_ops
        ds_deps += new_ds_deps
        # Return op deps, ds deps, parent ops, and updated arguments
        unique_parent_ops = []
        for op in parent_ops:
            if op not in unique_parent_ops:
                unique_parent_ops.append(op)
        return list(set(op_deps)), list(set(ds_deps)), unique_parent_ops, arguments

    @staticmethod
    def _replace_ds_args(arguments):
        """
        Traverses all args recursively and replaces instances of Datasets with fqtns and adds returns their dependencies
        """

        def replace_string(s):
            op_deps = []
            parent_ops = []
            ds_deps = []
            if isinstance(s, Dataset):
                op_deps.append(s.fqtn)
                if s._api_dataset:
                    s._assert_can_transform()
                    ds_deps.append(s._api_dataset.id)
                else:
                    parent_ops = s._operations
                s = s.fqtn
            return s, op_deps, parent_ops, ds_deps

        def replace_list(l):
            all_op_deps = []
            all_parent_ops = []
            all_ds_deps = []
            for i in range(len(l)):
                if isinstance(l[i], dict):
                    l[i], op_deps, parent_ops, ds_deps = replace_dict(l[i])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
                elif isinstance(l[i], list):
                    l[i], op_deps, parent_ops, ds_deps = replace_list(l[i])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
                else:
                    l[i], op_deps, parent_ops, ds_deps = replace_string(l[i])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
            return l, all_op_deps, all_parent_ops, all_ds_deps

        def replace_dict(d):
            all_op_deps = []
            all_parent_ops = []
            all_ds_deps = []
            for k in d.keys():
                if isinstance(d[k], dict):
                    d[k], op_deps, parent_ops, ds_deps = replace_dict(d[k])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
                if isinstance(d[k], list):
                    d[k], op_deps, parent_ops, ds_deps = replace_list(d[k])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
                else:
                    d[k], op_deps, parent_ops, ds_deps = replace_string(d[k])
                    all_op_deps += op_deps
                    all_parent_ops += parent_ops
                    all_ds_deps += ds_deps
            return d, all_op_deps, all_parent_ops, all_ds_deps

        return replace_dict(arguments)

    @staticmethod
    def _update_op_deps__ds_deps__and_parent_ops(
        ds_arg: Dataset, op_deps: List[str], parent_ops: List[OperationCreate], ds_deps: List[int]
    ) -> Tuple[List[str], List[OperationCreate], List[int]]:
        """
        Return updated DS args, deps, and op deps if arg to transform is type Dataset
        """
        op_deps = op_deps.copy()
        parent_ops = parent_ops.copy()
        ds_deps = ds_deps.copy()

        # Add input DS as operation dependency
        op_deps.append(ds_arg.fqtn)

        # If the dataset from the API, make sure is published
        # and has a respective DWTable. If so update dataset deps
        if ds_arg._api_dataset:
            ds_arg._assert_can_transform()
            ds_deps.append(ds_arg._api_dataset.id)
        # If it's an offline dataset, grab it's operations to set
        # as parents to output DS operations.
        # We need to make sure we aren't adding the same operation twice as well
        else:
            for ds_op in ds_arg._operations:
                if ds_op not in parent_ops:
                    parent_ops.append(ds_op)

        return op_deps, parent_ops, ds_deps


def _assert_value_types_of_args(
    transform: Transform,
    supplied_args: List[Dict[str, Union[str, int, List, Dict, Dataset]]],
    base_dataset: Dataset,
) -> None:
    """
    Raise exception if any of the supplied arguments are are not a valid type

    We especially need this to make sure args which expect datasets
    are ds objs and not fqtns, so we can add dependencies properly

    TODO: Add validation on other data types like `int_list`

    Args:
        transform: Transform applying
        supplied_args: Arguments of the supplied transform
        base_dataset: The base dataset which we're transforming
                      this is needed we can do `isinstance(arg, dataset.__class__)`
    """
    for transform_arg in transform.arguments:
        # We don't want to assert transform arguments
        # that are optional but not supplied, so skip
        # if not in user supplied args dict
        if transform_arg.name in supplied_args:
            supplied_arg_val = supplied_args[transform_arg.name]

            # If argument expects type 'table'
            # raise error if not of type 'Dataset'
            if transform_arg.type == TABLE_ARG_TYPE:
                if not isinstance(supplied_arg_val, base_dataset.__class__):
                    raise errors.ParameterValueError(
                        message=f"The {transform.name}() transform's parameter "
                        f"{transform_arg.name!r} requires the value to be "
                        f"a single Dataset obj. Got {type(supplied_arg_val)} instead"
                    )

            # If argument expects type 'table_list'
            # raise error if not of type List['Dataset']
            elif transform_arg.type == TABLE_LIST_ARG_TYPE:
                if not isinstance(supplied_arg_val, list) or any(
                    [not isinstance(x, base_dataset.__class__) for x in supplied_arg_val]
                ):
                    raise errors.ParameterValueError(
                        message=f"The {transform.name}() transform's parameter "
                        f"{transform_arg.name!r} requires the value to "
                        "be a non-empty list of Dataset objs."
                    )


def _get_transforms() -> List[Transform]:
    """
    Get and set available transforms from the API to be used
    directly as functions of Dataset if not retrieved yet
    """
    # Get available transforms from the API to be used directly as functions of Dataset
    from pyrasgo.api import Get

    try:
        return Get().transforms()
    except Exception:
        print('Unable to fetch available transforms from Rasgo.  Will not be able to transform this Dataset')
        return []


def _gen_func_signature(func: Callable, transform: Transform) -> inspect.Signature:
    """
    Creates and returns a transform param signature.

    This is shown documentation for the parameters when hitting shift tab in a notebook
    """
    # Get current signature of function
    sig = inspect.signature(func)

    # Create Signature Params for Transform Args
    transform_params = []
    for t_arg in transform.arguments:
        p = Parameter(name=t_arg.name, kind=Parameter.KEYWORD_ONLY)
        transform_params.append(p)

    # Add `operation_name` param as last in signature with type annotation
    op_name_param = Parameter(
        name='operation_name',
        kind=Parameter.KEYWORD_ONLY,
        annotation=Optional[str],
        default=None,
    )
    transform_params.append(op_name_param)

    # Return new signature
    return sig.replace(parameters=transform_params)


def _gen_func_docstring(transform: Transform) -> str:
    """
    Generate and return a docstring for a transform func
    with transform description, args, and return specified.
    """
    # Have start of docstring be transform description
    docstring = f"\n{transform.description}"

    # Add transform args to func docstring
    docstring = f"{docstring}\n  Args:"
    for t_arg in transform.arguments:
        docstring = f"{docstring}\n    {t_arg.name}: {t_arg.description}"
    docstring = f"{docstring}\n    operation_name: Name to set for the operation"

    # Add return to docstring
    docstring = (
        f"{docstring}\n\n  Returns:\n    Returns an new dataset with the referenced "
        f"{transform.name!r} added to this dataset's definition"
    )
    return docstring



from typing import Optional, List

from pydantic import BaseModel, Field


class DataColumn(BaseModel):
    """
    Contract to nest in a DataTable
    """

    id: int
    column_name: str = Field(alias='columnName')
    data_type: str = Field(alias='dataType')

    class Config:
        allow_population_by_field_name = True


class DataTable(BaseModel):
    """
    Contract to nest in a Dataset
    """

    id: int
    table_name: str = Field(alias='tableName')
    database_name: str = Field(alias='databaseName')
    schema_name: str = Field(alias='schemaName')
    fqtn: Optional[str]
    table_type: str = Field(alias='tableType')

    class Config:
        allow_population_by_field_name = True


class DataTableWithColumns(DataTable):
    """
    Contract to nest in an OperationSet
    """

    columns: Optional[List[DataColumn]]



from .accelerators import (
    AcceleratorOperationCreate,
    AcceleratorArgumentCreate,
    AcceleratorCreate,
    Accelerator,
    AcceleratorApply,
)
from .attributes import Attribute
from .dataset import Dataset, DatasetBulk, DatasetCreate, DatasetUpdate, DatasetSourceType
from .dataset_column import DatasetColumn, DatasetColumnUpdate
from .dw_operation import Operation, OperationCreate
from .dw_operation_set import (
    OperationSet,
    OperationSetAsyncEvent,
    OperationSetAsyncTask,
    OperationSetCreate,
    OperationSetOfflineAsyncEvent,
    OperationSetOfflineAsyncTask,
)
from .dw_table import DataColumn, DataTable, DataTableWithColumns
from .enums import DataType
from .organization import Organization
from .stats import GenerateStat
from .status_tracking import StatusTracking
from .transform import (
    Transform,
    TransformArgument,
    TransformArgumentCreate,
    TransformCreate,
    TransformExecute,
    TransformUpdate,
)
from .user import User
from .insight import InsightCreate, Insight
from .metric import Metric, MetricCreate, MetricUpdate, TimeGrain, Filter
from .offline import OfflineDataset
